# Copyright (c) 2013 Hortonworks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from savanna.openstack.common import jsonutils as json
from savanna.plugins.hdp import configprovider as cfg


class ClusterSpec():
    def _get_servers_from_savanna_cluster(self, cluster):
        servers = []
        for node_group in cluster.node_groups:
            for server in node_group.instances:
                setattr(server, 'role', node_group.name)
                setattr(server, 'node_processes', node_group.node_processes)
                servers.append(server)

        return servers

    def __init__(self, cluster_template, cluster=None):
        self.services = []
        self.configurations = {}
        self.node_groups = {}
        self.str = cluster_template

        servers = []
        if cluster is not None:
            if hasattr(cluster, 'node_groups'):
                servers = self._get_servers_from_savanna_cluster(cluster)
            else:
                servers = cluster.instances

            host_manifest = self._generate_host_manifest(servers)
            #TODO(jspeidel): don't hard code ambari server
            ambari_server = self._get_ambari_host(servers)
            if ambari_server is not None:
                cluster_template = cluster_template.replace('%AMBARI_HOST%',
                                                            ambari_server.fqdn)
            else:
                raise RuntimeError('No Ambari server host found')

            self.str = self._add_manifest_to_config(cluster_template,
                                                    host_manifest)

        template_json = json.loads(self.str)
        self._parse_services(template_json)
        self._parse_configurations(template_json)
        self._parse_host_component_mappings(template_json)

    def _get_ambari_host(self, servers):
        # iterate thru servers and find the master server
        host = next((server for server in servers
                     if server.node_processes is not None and
                     'AMBARI_SERVER' in server.node_processes), None)
        if host is None:
            host = next((server for server in servers
                         if server.role == 'MASTER'), None)
        return host

    def normalize(self):
        return NormalizedClusterConfig(self)

    def _parse_services(self, template_json):
        for s in template_json['services']:
            service = Service(s['name'])

            self.services.append(service)
            for c in s['components']:
                component = Component(c['name'], c['type'], c['cardinality'])
                service.add_component(component)

            configs = self._parse_configurations(s)
            for config in configs:
                service.add_configuration(config)

    def _parse_configurations(self, template_json):
        config_names = []
        for config in template_json['configurations']:
            config_props = {}
            name = config['name']
            config_names.append(name)
            if name in self.configurations:
                config_props = self.configurations[name]
            else:
                self.configurations[name] = config_props

            if 'properties' in config:
                for prop in config['properties']:
                    config_props[prop['name']] = prop['value']

        return config_names

    def _parse_host_component_mappings(self, template_json):
        for group in template_json['host_role_mappings']:
            node_group = NodeGroup(group['name'])
            for component in group['components']:
                node_group.add_component(component['name'])
            for host in group['hosts']:
                if 'predicate' in host:
                    node_group.predicate = host['predicate']
                if 'cardinality' in host:
                    node_group.cardinality = host['cardinality']
                if 'default_count' in host:
                    node_group.default_count = host['default_count']
            self.node_groups[node_group.name] = node_group

    def _generate_host_manifest(self, servers):
        host_manifest = {}
        hosts = []
        host_id = 1

        for server in servers:
            hosts.append({'host_id': host_id,
                          'hostname': server.hostname,
                          'role': server.role,
                          'vm_image': server.nova_info.image,
                          'vm_flavor': server.nova_info.flavor,
                          'public_ip': server.management_ip,
                          'private_ip': server.internal_ip})
            host_id += 1

        host_manifest['hosts'] = hosts
        return json.dumps(host_manifest).strip('{}')

    def _add_manifest_to_config(self, cluster_template, host_manifest):
        # add the host manifest to the enf of the cluster template

        return '{0},\n{1}\n}}'.format(cluster_template.rstrip('}'),
                                      host_manifest)


class Service():
    def __init__(self, name):
        self.name = name
        self.configurations = []
        self.components = []

    def add_component(self, component):
        self.components.append(component)

    def add_configuration(self, configuration):
        self.configurations.append(configuration)


class Component():
    def __init__(self, name, component_type, cardinality):
        self.name = name
        self.type = component_type
        self.cardinality = cardinality


class NodeGroup():
    def __init__(self, name):
        self.name = name
        self.components = []
        self.predicate = None
        self.cardinality = None
        self.default_count = None

    def add_component(self, component):
        self.components.append(component)


class NormalizedClusterConfig():
    def __init__(self, cluster_spec):
        #TODO(jspeidel): get from stack config
        self.hadoop_version = '1.3.0'
        self.cluster_configs = []
        self.node_groups = []
        self.config = cfg.ConfigurationProvider(
            json.load(open(os.path.join(os.path.dirname(__file__), 'resources',
                                        'ambari-config-resource.json'), "r")))

        self._parse_configurations(cluster_spec.configurations)
        self._parse_node_groups(cluster_spec.node_groups)

    def _parse_configurations(self, configurations):
        for config_name, properties in configurations.items():
            for prop, value in properties.items():
                target = self._get_property_target(config_name, prop)
                prop_type = self._get_property_type(prop, value)
                #todo: should we supply a scope?
                self.cluster_configs.append(
                    NormalizedConfigEntry(NormalizedConfig(
                        prop, prop_type, value, target, 'cluster'), value))

    def _parse_node_groups(self, node_groups):
        for node_group in node_groups.values():
            self.node_groups.append(NormalizedNodeGroup(node_group))

    def _get_property_target(self, config, prop):
        # Once config resource is complete we won't need to fall through
        # based on config type
        target = self.config.get_applicable_target(prop)
        if not target:
            if config == 'hdfs-site':
                target = 'service:HDFS'
            elif config == 'mapred-site':
                target = 'service:MAPREDUCE'
            else:
                target = 'general'
        else:
            if target != 'general':
                target = "service:" + target

        return target

    def _get_property_type(self, prop, value):
        #TODO(jspeidel): seems that all numeric prop values in default config
        # are encoded as strings.  This may be incorrect.
        #TODO(jspeidel): should probably analyze string value to determine if
        # it is numeric
        #TODO(jspeidel): would then need to know whether Ambari expects a
        # string or a numeric value
        prop_type = type(value).__name__
        #print 'Type: {0}'.format(prop_type)
        if prop_type == 'str' or prop_type == 'unicode' or value == '':
            return 'string'
        elif prop_type == 'int':
            return 'integer'
        elif prop_type == 'bool':
            return 'boolean'
        else:
            raise ValueError(
                "Could not determine property type for property '{0}' with "
                "value: {1}".
                format(prop, value))


class NormalizedConfig():
    def __init__(self, name, config_type, default_value, target, scope):
        self.name = name
        self.description = None
        self.type = config_type
        self.default_value = default_value
        self.is_optional = False
        self.applicable_target = target
        self.scope = scope


class NormalizedConfigEntry():
    def __init__(self, config, value):
        self.config = config
        self.value = value


class NormalizedNodeGroup():
    def __init__(self, node_group):
        self.name = node_group.name
        self.node_processes = node_group.components
        self.node_configs = None
        #TODO(jpseidel): should not have to specify img/flavor
        self.img = None
        self.flavor = None
        self.count = node_group.default_count
        #TODO(jspeidel): self.requirements



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna import context
from savanna.db import models as m
from savanna.db import storage as s
from savanna.openstack.common import excutils
from savanna.openstack.common import log as logging
from savanna.openstack.common import uuidutils
from savanna.plugins import base as plugin_base
from savanna.plugins import provisioning
from savanna.service import instances as i
from savanna.utils.openstack import nova

LOG = logging.getLogger(__name__)


## Cluster ops

get_clusters = s.get_clusters
get_cluster = s.get_cluster


def scale_cluster(cluster_id, data):
    cluster = get_cluster(id=cluster_id)
    plugin = plugin_base.PLUGINS.get_plugin(cluster.plugin_name)
    existing_node_groups = data.get('resize_node_groups', [])
    additional_node_groups = data.get('add_node_groups', [])

    #the next map is the main object we will work with
    #to_be_enlarged : {node_group_name: desired_amount_of_instances}
    to_be_enlarged = {}
    for ng in existing_node_groups:
        to_be_enlarged.update({ng['name']: ng['count']})

    additional = construct_ngs_for_scaling(additional_node_groups)

    try:
        context.model_update(cluster, status='Validating')
        plugin.validate_scaling(cluster, to_be_enlarged, additional)
    except Exception:
        with excutils.save_and_reraise_exception():
            context.model_update(cluster, status='Active')

    # If we are here validation is successful.
    # So let's update bd and to_be_enlarged map:
    for add_n_g in additional:
        cluster.node_groups.append(add_n_g)
        to_be_enlarged.update({add_n_g.name: additional[add_n_g]})
    context.model_save(cluster)

    context.spawn(_provision_nodes, cluster_id, to_be_enlarged)
    return cluster


def create_cluster(values):
    cluster = s.create_cluster(values)
    plugin = plugin_base.PLUGINS.get_plugin(cluster.plugin_name)

    # validating cluster
    try:
        context.model_update(cluster, status='Validating')
        plugin.validate(cluster)
    except Exception as ex:
        with excutils.save_and_reraise_exception():
            context.model_update(cluster, status='Error',
                                 status_description=str(ex))

    context.spawn(_provision_cluster, cluster.id)

    return cluster


#node_group_names_map = {node_group_name:desired_amount_of_instances}
def _provision_nodes(cluster_id, node_group_names_map):
    cluster = get_cluster(id=cluster_id)
    plugin = plugin_base.PLUGINS.get_plugin(cluster.plugin_name)

    context.model_update(cluster, status='Scaling')
    instances = i.scale_cluster(cluster, node_group_names_map, plugin)

    if instances:
        context.model_update(cluster, status='Configuring')
        plugin.scale_cluster(cluster, instances)

    # cluster is now up and ready
    context.model_update(cluster, status='Active')


def _provision_cluster(cluster_id):
    cluster = get_cluster(id=cluster_id)
    plugin = plugin_base.PLUGINS.get_plugin(cluster.plugin_name)

    # updating cluster infra
    context.model_update(cluster, status='InfraUpdating')
    plugin.update_infra(cluster)

    # creating instances and configuring them
    i.create_cluster(cluster)

    # configure cluster
    context.model_update(cluster, status='Configuring')
    plugin.configure_cluster(cluster)

    # starting prepared and configured cluster
    context.model_update(cluster, status='Starting')
    plugin.start_cluster(cluster)

    # cluster is now up and ready
    context.model_update(cluster, status='Active')

    return cluster


def terminate_cluster(**args):
    cluster = get_cluster(**args)
    context.model_update(cluster, status='Deleting')

    plugin = plugin_base.PLUGINS.get_plugin(cluster.plugin_name)
    plugin.on_terminate_cluster(cluster)

    i.shutdown_cluster(cluster)
    s.terminate_cluster(cluster)


## ClusterTemplate ops

get_cluster_templates = s.get_cluster_templates
get_cluster_template = s.get_cluster_template
create_cluster_template = s.create_cluster_template
terminate_cluster_template = s.terminate_cluster_template


## NodeGroupTemplate ops

get_node_group_templates = s.get_node_group_templates
get_node_group_template = s.get_node_group_template
create_node_group_template = s.create_node_group_template
terminate_node_group_template = s.terminate_node_group_template


## Plugins ops

def get_plugins():
    return plugin_base.PLUGINS.get_plugins(
        base=provisioning.ProvisioningPluginBase)


def get_plugin(plugin_name, version=None):
    plugin = plugin_base.PLUGINS.get_plugin(plugin_name)
    if plugin:
        res = plugin.as_resource()
        if version:
            if version in plugin.get_versions():
                configs = plugin.get_configs(version)
                res._info['configs'] = [c.dict for c in configs]
                processes = plugin.get_node_processes(version)
                res._info['node_processes'] = processes
                required_image_tags = plugin.get_required_image_tags(version)
                res._info['required_image_tags'] = required_image_tags
            else:
                return None
        return res


def convert_to_cluster_template(plugin_name, version, config_file):
    plugin = plugin_base.PLUGINS.get_plugin(plugin_name)
    tenant_id = context.current().tenant_id
    name = uuidutils.generate_uuid()
    ct = m.ClusterTemplate(name, tenant_id, plugin_name, version)
    plugin.convert(ct, config_file)

    return s.persist_cluster_template(ct)


def construct_ngs_for_scaling(additional_node_groups):
    additional = {}
    for ng in additional_node_groups:
        tmpl_id = ng.get('node_group_template_id')
        count = ng['count']
        if tmpl_id:
            tmpl = get_node_group_template(id=tmpl_id)
            node_group = tmpl.to_object(ng, m.NodeGroup)
        else:
            node_group = m.NodeGroup(**ng)
            #need to set 0 because tmpl.to_object overwrote count
        node_group.count = 0
        additional.update({node_group: count})
    return additional

## Image Registry


def get_images(tags):
    return nova.client().images.list_registered(tags)


def get_image(**kwargs):
    if len(kwargs) == 1 and 'id' in kwargs:
        return nova.client().images.get(kwargs['id'])
    else:
        return nova.client().images.find(**kwargs)


def register_image(image_id, username, description=None):
    client = nova.client()
    client.images.set_description(image_id, username, description)
    return client.images.get(image_id)


def unregister_image(image_id):
    client = nova.client()
    client.images.unset_description(image_id)
    return client.images.get(image_id)


def add_image_tags(image_id, tags):
    client = nova.client()
    client.images.tag(image_id, tags)
    return client.images.get(image_id)


def remove_image_tags(image_id, tags):
    client = nova.client()
    client.images.untag(image_id, tags)
    return client.images.get(image_id)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pkg_resources as pkg
import re
import xml.dom.minidom as xml

from savanna import version


def load_xml_document(file_name):
    return xml.parse(pkg.resource_filename(
        version.version_info.package, file_name))


def load_hadoop_xml_defaults(file_name):
    doc = load_xml_document(file_name)
    configs = []
    prop = doc.getElementsByTagName('property')
    for elements in prop:
        configs.append({
            "name": _get_text_from_node(elements, 'name'),
            "value": _get_text_from_node(elements, 'value'),
            "description": _adjust_description(
                _get_text_from_node(elements, 'description'))
        })
    return configs


def _get_text_from_node(element, name):
    element = element.getElementsByTagName(name)
    return element[0].firstChild.nodeValue if (
        element and element[0].hasChildNodes()) else ''


def _adjust_description(text):
    return re.sub(r"\n *|\t", "", str(text))


def create_hadoop_xml(configs, global_conf):
    doc = xml.Document()

    pi = doc.createProcessingInstruction('xml-stylesheet',
                                         'type="text/xsl" '
                                         'href="configuration.xsl"')
    doc.insertBefore(pi, doc.firstChild)

    # Create the <configuration> base element
    configuration = doc.createElement("configuration")
    doc.appendChild(configuration)

    for name, value in configs.items():
        if name in [cfg['name'] for cfg in global_conf]:
            # Create the <property> element
            xml_prop = doc.createElement("property")
            configuration.appendChild(xml_prop)

            # Create a <name> element in <property>
            name_element = doc.createElement("name")
            xml_prop.appendChild(name_element)

            # Give the <name> element some hadoop config name
            name_text = doc.createTextNode(str(name))
            name_element.appendChild(name_text)

            # Create a <value> element in <property>
            value_element = doc.createElement("value")
            xml_prop.appendChild(value_element)

            # Give the <value> element some hadoop config value
            value_text = doc.createTextNode(str(value))
            value_element.appendChild(value_text)

    # Return newly created XML
    return doc.toprettyxml(indent="  ")



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest2

from savanna.db import models as m


class VolumesSupportTest(unittest2.TestCase):
    def test_instance_storage_paths_wo_volumes(self):
        i = m.NodeGroup('ng-1', 'f-1', [], 1)

        self.assertEqual(
            [
                '/mnt',
            ],
            i.storage_paths)

    def test_instance_storage_paths(self):
        i = m.NodeGroup('ng-1', 'f-1', [], 1, volumes_per_node=3)

        self.assertEqual(
            [
                '/volumes/disk1',
                '/volumes/disk2',
                '/volumes/disk3',
            ],
            i.storage_paths)

    def test_instance_storage_paths_custem_prefix(self):
        i = m.NodeGroup('ng-1', 'f-1', [], 1,
                        volumes_per_node=3,
                        volume_mount_prefix='/foo')

        self.assertEqual(
            [
                '/foo1',
                '/foo2',
                '/foo3',
            ],
            i.storage_paths)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""v02_release

Revision ID: 5a3671e1e2de
Revises: None
Create Date: 2013-07-09 08:18:21.330480

"""

# revision identifiers, used by Alembic.
revision = '5a3671e1e2de'
down_revision = None

from alembic import op
import sqlalchemy as sa

from savanna.utils import sqlatypes as st

sa.JSONEncoded = st.JSONEncoded


def upgrade():
    op.create_table(
        'NodeGroupTemplate',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('description', sa.String(length=200), nullable=True),
        sa.Column('tenant_id', sa.String(length=36), nullable=True),
        sa.Column('plugin_name', sa.String(length=80), nullable=False),
        sa.Column('hadoop_version', sa.String(length=80), nullable=False),
        sa.Column('name', sa.String(length=80), nullable=False),
        sa.Column('flavor_id', sa.String(length=36), nullable=False),
        sa.Column('image_id', sa.String(length=36), nullable=True),
        sa.Column('node_processes', sa.JSONEncoded(), nullable=True),
        sa.Column('node_configs', sa.JSONEncoded(), nullable=True),
        sa.Column('volumes_per_node', sa.Integer(), nullable=True),
        sa.Column('volumes_size', sa.Integer(), nullable=True),
        sa.Column('volume_mount_prefix', sa.String(length=80), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'tenant_id')
    )
    op.create_table(
        'ClusterTemplate',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('description', sa.String(length=200), nullable=True),
        sa.Column('tenant_id', sa.String(length=36), nullable=True),
        sa.Column('plugin_name', sa.String(length=80), nullable=False),
        sa.Column('hadoop_version', sa.String(length=80), nullable=False),
        sa.Column('extra', sa.JSONEncoded(), nullable=True),
        sa.Column('name', sa.String(length=80), nullable=False),
        sa.Column('cluster_configs', sa.JSONEncoded(), nullable=True),
        sa.Column('default_image_id', sa.String(length=36), nullable=True),
        sa.Column('anti_affinity', sa.JSONEncoded(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'tenant_id')
    )
    op.create_table(
        'TemplatesRelation',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=80), nullable=False),
        sa.Column('flavor_id', sa.String(length=36), nullable=False),
        sa.Column('image_id', sa.String(length=36), nullable=True),
        sa.Column('node_processes', sa.JSONEncoded(), nullable=True),
        sa.Column('node_configs', sa.JSONEncoded(), nullable=True),
        sa.Column('volumes_per_node', sa.Integer(), nullable=True),
        sa.Column('volumes_size', sa.Integer(), nullable=True),
        sa.Column('volume_mount_prefix', sa.String(length=80), nullable=True),
        sa.Column('count', sa.Integer(), nullable=False),
        sa.Column('cluster_template_id', sa.String(length=36), nullable=True),
        sa.Column('node_group_template_id', sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['cluster_template_id'],
                                ['ClusterTemplate.id'], ),
        sa.ForeignKeyConstraint(['node_group_template_id'],
                                ['NodeGroupTemplate.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'Cluster',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('description', sa.String(length=200), nullable=True),
        sa.Column('tenant_id', sa.String(length=36), nullable=True),
        sa.Column('plugin_name', sa.String(length=80), nullable=False),
        sa.Column('hadoop_version', sa.String(length=80), nullable=False),
        sa.Column('extra', sa.JSONEncoded(), nullable=True),
        sa.Column('name', sa.String(length=80), nullable=False),
        sa.Column('cluster_configs', sa.JSONEncoded(), nullable=True),
        sa.Column('default_image_id', sa.String(length=36), nullable=True),
        sa.Column('anti_affinity', sa.JSONEncoded(), nullable=True),
        sa.Column('status', sa.String(length=80), nullable=True),
        sa.Column('status_description', sa.String(length=200), nullable=True),
        sa.Column('private_key', sa.Text(), nullable=True),
        sa.Column('user_keypair_id', sa.String(length=80), nullable=True),
        sa.Column('info', sa.JSONEncoded(), nullable=True),
        sa.Column('cluster_template_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['cluster_template_id'],
                                ['ClusterTemplate.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'tenant_id')
    )
    op.create_table(
        'NodeGroup',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=80), nullable=False),
        sa.Column('flavor_id', sa.String(length=36), nullable=False),
        sa.Column('image_id', sa.String(length=36), nullable=True),
        sa.Column('node_processes', sa.JSONEncoded(), nullable=True),
        sa.Column('node_configs', sa.JSONEncoded(), nullable=True),
        sa.Column('volumes_per_node', sa.Integer(), nullable=True),
        sa.Column('volumes_size', sa.Integer(), nullable=True),
        sa.Column('volume_mount_prefix', sa.String(length=80), nullable=True),
        sa.Column('count', sa.Integer(), nullable=False),
        sa.Column('cluster_id', sa.String(length=36), nullable=True),
        sa.Column('node_group_template_id', sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['cluster_id'], ['Cluster.id'], ),
        sa.ForeignKeyConstraint(['node_group_template_id'],
                                ['NodeGroupTemplate.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'cluster_id')
    )
    op.create_table(
        'Instance',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('extra', sa.JSONEncoded(), nullable=True),
        sa.Column('node_group_id', sa.String(length=36), nullable=True),
        sa.Column('instance_id', sa.String(length=36), nullable=False),
        sa.Column('instance_name', sa.String(length=80), nullable=False),
        sa.Column('internal_ip', sa.String(length=15), nullable=True),
        sa.Column('management_ip', sa.String(length=15), nullable=True),
        sa.Column('volumes', sa.JSONEncoded(), nullable=True),
        sa.ForeignKeyConstraint(['node_group_id'], ['NodeGroup.id'], ),
        sa.PrimaryKeyConstraint('instance_id'),
        sa.UniqueConstraint('instance_id', 'node_group_id')
    )


def downgrade():
    op.drop_table('Instance')
    op.drop_table('NodeGroup')
    op.drop_table('Cluster')
    op.drop_table('TemplatesRelation')
    op.drop_table('ClusterTemplate')
    op.drop_table('NodeGroupTemplate')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from novaclient.v1_1 import images


PROP_DESCR = '_savanna_description'
PROP_USERNAME = '_savanna_username'
PROP_TAG = '_savanna_tag_'


def _iter_tags(meta):
    for key in meta:
        if key.startswith(PROP_TAG) and meta[key]:
            yield key[len(PROP_TAG):]


def _ensure_tags(tags):
    if not tags:
        return []
    return [tags] if type(tags) in [str, unicode] else tags


class SavannaImage(images.Image):
    def __init__(self, manager, info, loaded=False):
        info['description'] = info.get('metadata', {}).get(PROP_DESCR)
        info['username'] = info.get('metadata', {}).get(PROP_USERNAME)
        info['tags'] = [tag for tag in _iter_tags(info.get('metadata', {}))]
        super(SavannaImage, self).__init__(manager, info, loaded)

    def tag(self, tags):
        self.manager.tag(self, tags)

    def untag(self, tags):
        self.manager.untag(self, tags)

    def set_description(self, username, description=None):
        self.manager.set_description(self, username, description)

    def unset_description(self):
        self.manager.unset_description(self)

    @property
    def dict(self):
        return self.to_dict()

    @property
    def wrapped_dict(self):
        return {'image': self.dict}

    def to_dict(self):
        result = self._info.copy()
        del result['links']
        return result


class SavannaImageManager(images.ImageManager):
    """Manage :class:`SavannaImage` resources.

    This is an extended version of nova client's ImageManager with support of
    additional description and image tags stored in images' meta.
    """
    resource_class = SavannaImage

    def set_description(self, image, username, description=None):
        """Sets human-readable information for image.

        For example:

            Ubuntu 13.04 x64 with Java 1.7u21 and Apache Hadoop 1.1.1, ubuntu
        """
        self.set_meta(image, {
            PROP_DESCR: description,
            PROP_USERNAME: username,
        })

    def unset_description(self, image):
        """Unsets all Savanna-related information.

        It removes username, description and tags from the specified image.
        """
        image = self.get(image)
        meta = [PROP_TAG + tag for tag in image.tags]
        if image.description is not None:
            meta += [PROP_DESCR]
        if image.username is not None:
            meta += [PROP_USERNAME]
        self.delete_meta(image, meta)

    def tag(self, image, tags):
        """Adds tags to the specified image."""
        tags = _ensure_tags(tags)

        self.set_meta(image, dict((PROP_TAG + tag, True) for tag in tags))

    def untag(self, image, tags):
        """Removes tags from the specified image."""
        tags = _ensure_tags(tags)

        self.delete_meta(image, [PROP_TAG + tag for tag in tags])

    def list_by_tags(self, tags):
        """Returns images having all of the specified tags."""
        tags = _ensure_tags(tags)
        return [i for i in self.list() if set(tags).issubset(i.tags)]

    def list_registered(self, tags=None):
        tags = _ensure_tags(tags)
        return [i for i in self.list()
                if i.username and set(tags).issubset(i.tags)]



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def get_node_groups(cluster, proc_list=list()):
    proc_list = [proc_list] if type(proc_list) in [str, unicode] else proc_list
    return [ng for ng in cluster.node_groups
            if set(proc_list).issubset(ng.node_processes)]


def get_instances(cluster, proc_list=list()):
    nodes = get_node_groups(cluster, proc_list)
    return reduce(lambda a, b: a + b.instances, nodes, [])


def get_namenode(cluster):
    nn = get_instances(cluster, "namenode")
    return nn[0] if nn else None


def get_jobtracker(cluster):
    jt = get_instances(cluster, "jobtracker")
    return jt[0] if jt else None


def get_datanodes(cluster):
    return get_instances(cluster, 'datanode')


def get_tasktrackers(cluster):
    return get_instances(cluster, 'tasktracker')


def get_secondarynamenodes(cluster):
    return get_instances(cluster, 'secondarynamenode')


def generate_host_names(nodes):
    return "\n".join([n.hostname for n in nodes])


def generate_fqdn_host_names(nodes):
    return "\n".join([n.fqdn for n in nodes])



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from oslo.config import cfg

from savanna import context
from savanna.utils import xmlutils as x


LOG = logging.getLogger(__name__)
CONF = cfg.CONF
HADOOP_SWIFT_AUTH_URL = 'fs.swift.service.savanna.auth.url'
HADOOP_SWIFT_TENANT = 'fs.swift.service.savanna.tenant'


def _retrieve_auth_url():
    url = "{0}://{1}:{2}/v2.0/tokens/".format(
        CONF.os_auth_protocol, CONF.os_auth_host, CONF.os_auth_port)
    return url


def _retrieve_tenant():
    try:
        return context.current().headers['X-Tenant-Name']
    except RuntimeError:
        LOG.error("Cannot retrieve tenant for swift integration. "
                  "Stop cluster creation")
        #todo(slukjanov?) raise special error here
        raise RuntimeError("Cannot retrieve tenant for swift integration")


def get_swift_configs():
    configs = x.load_hadoop_xml_defaults('swift/resources/conf-template.xml')
    for conf in configs:
        if conf['name'] == HADOOP_SWIFT_AUTH_URL:
            conf['value'] = _retrieve_auth_url()
        if conf['name'] == HADOOP_SWIFT_TENANT:
            conf['value'] = _retrieve_tenant()

    result = [cfg for cfg in configs if cfg['value']]
    LOG.info("Swift would be integrated with the following "
             "params: %s", result)
    return result



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna.service import api
from savanna.service.validations import clusters as c
from savanna.tests.unit.service.validation import utils as u


class TestClusterCreateValidation(u.ValidationTestCase):
    def setUp(self):
        self._create_object_fun = c.check_cluster_create
        self.scheme = c.CLUSTER_SCHEMA
        api.plugin_base.setup_plugins()

    def test_cluster_create_v_plugin_vers(self):
        self._assert_create_object_validation(
            data={
                'name': 'testname',
                'plugin_name': 'vanilla',
                'hadoop_version': '1'
            },
            bad_req_i=(1, "INVALID_REFERENCE",
                       "Requested plugin 'vanilla' "
                       "doesn't support version '1'"),
        )

    def test_cluster_create_v_required(self):
        self._assert_create_object_validation(
            data={},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'name' is a required property")
        )
        self._assert_create_object_validation(
            data={
                'name': 'test-name'
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'plugin_name' is a required property")
        )
        self._assert_create_object_validation(
            data={
                'name': 'testname',
                'plugin_name': 'vanilla'
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'hadoop_version' is a required property")
        )

    def test_cluster_create_v_types(self):
        data = {
            'name': "testname",
            'plugin_name': "vanilla",
            'hadoop_version': "1.1.2"
        }
        self._assert_types(data)

    def test_cluster_create_v_name_base(self):
        data = {
            'name': "testname",
            'plugin_name': "vanilla",
            'hadoop_version': "1.1.2"
        }
        self._assert_object_name_validation(data, 'hostname')

    def test_cluster_create_v_unique_cl(self):
        data = {
            'name': 'test',
            'plugin_name': 'vanilla',
            'hadoop_version': '1.1.2'
        }
        self._assert_create_object_validation(
            data=data,
            bad_req_i=(1, 'NAME_ALREADY_EXISTS',
                       "Cluster with name 'test' already exists")
        )

    def test_cluster_create_v_keypair_exists(self):
        self._assert_create_object_validation(
            data={
                'name': "testname",
                'plugin_name': "vanilla",
                'hadoop_version': "1.1.2",
                'user_keypair_id': 'wrong_keypair'
            },
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Requested keypair 'wrong_keypair' not found")
        )

    def test_cluster_create_v_keypair_type(self):
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "vanilla",
                'hadoop_version': "1.1.2",
                'user_keypair_id': '1'
            },
            bad_req_i=(1, 'VALIDATION_ERROR',
                       "'1' is not a 'valid_name'")
        )
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "vanilla",
                'hadoop_version': "1.1.2",
                'user_keypair_id': '!'},
            bad_req_i=(1, 'VALIDATION_ERROR',
                       "'!' is not a 'valid_name'")
        )

    def test_cluster_create_v_image_exists(self):
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "vanilla",
                'hadoop_version': "1.1.2",
                'default_image_id': '550e8400-e29b-41d4-a616-446655440000'
            },
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Requested image '550e8400-e29b-41d4-a616-446655440000'"
                       " is not registered")
        )

    def test_cluster_create_v_plugin_name_exists(self):
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "wrong_plugin",
                'hadoop_version': "1.1.2",
            },
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Savanna doesn't contain plugin "
                       "with name 'wrong_plugin'")
        )

    def test_cluster_create_v_cluster_configs(self):
        self._assert_cluster_configs_validation()

    def test_cluster_create_v_right_data(self):
        self._assert_create_object_validation(
            data={
                'name': "testname",
                'plugin_name': "vanilla",
                'hadoop_version': "1.1.2",
                'user_keypair_id': 'test_keypair',
                'cluster_configs': {
                    'HDFS': {
                        u'hadoop.tmp.dir': '/temp/'
                    }
                },
                'default_image_id': '550e8400-e29b-41d4-a716-446655440000'
            }
        )

    def test_cluster_create_v_default_image_required_tags(self):
        self._assert_cluster_default_image_tags_validation()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import sqlalchemy as sa
from sqlalchemy.ext import declarative
from sqlalchemy import orm

from savanna.openstack.common import timeutils
from savanna.openstack.common import uuidutils
from savanna.utils import resources
from savanna.utils import sqlatypes as st


class _SavannaBase(resources.BaseResource):
    """Base class for all Savanna Models."""

    created = sa.Column(sa.DateTime, default=timeutils.utcnow,
                        nullable=False)
    updated = sa.Column(sa.DateTime, default=timeutils.utcnow,
                        nullable=False, onupdate=timeutils.utcnow)

    __protected_attributes__ = ["created", "updated"]

    @declarative.declared_attr
    def __tablename__(cls):
        # Table name is equals to the class name
        return cls.__name__

    @property
    def __resource_name__(self):
        # convert CamelCase class name to camel_case
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', self.__class__.__name__)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def __iter__(self):
        self._i = iter(orm.object_mapper(self).columns)
        return self

    def next(self):
        n = self._i.next().name
        return n, getattr(self, n)

    def update(self, values):
        """Make the model object behave like a dict."""
        for k, v in values.iteritems():
            setattr(self, k, v)

    def iteritems(self):
        """Make the model object behave like a dict.

        Includes attributes from joins.
        """
        local = dict(self)
        joined = dict([(k, v) for k, v in self.__dict__.iteritems()
                       if not k[0] == '_'])
        local.update(joined)
        return local.iteritems()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def __repr__(self):
        """sqlalchemy based automatic __repr__ method."""
        items = ['%s=%r' % (col.name, getattr(self, col.name))
                 for col in self.__table__.columns]
        return "<%s.%s[object at %x] {%s}>" % (self.__class__.__module__,
                                               self.__class__.__name__,
                                               id(self), ', '.join(items))

    def to_dict(self):
        """sqlalchemy based automatic to_dict method."""
        d = {}
        for col in self.__table__.columns:
            if self._filter_field(col.name):
                continue
            d[col.name] = getattr(self, col.name)
        return d


SavannaBase = declarative.declarative_base(cls=_SavannaBase)


def _generate_unicode_uuid():
    return unicode(uuidutils.generate_uuid())


class IdMixin(object):
    """Id mixin, add to subclasses that have an id."""

    id = sa.Column(sa.String(36),
                   primary_key=True,
                   default=_generate_unicode_uuid)


class DescriptionMixin(object):
    """Description mixin, add to subclasses that have a description."""

    description = sa.Column(sa.String(200))


class TenantMixin(object):
    """Tenant mixin, add to subclasses that have a tenant."""

    __filter_cols__ = ['tenant_id']

    tenant_id = sa.Column(sa.String(36))


class PluginSpecificMixin(object):
    """Plugin specific info mixin, add to subclass that plugin specific."""

    plugin_name = sa.Column(sa.String(80), nullable=False)
    hadoop_version = sa.Column(sa.String(80), nullable=False)


class ExtraMixin(object):
    """Extra info mixin, add to subclass that stores extra data w/o schema."""

    __filter_cols__ = ['extra']

    extra = sa.Column(st.JsonDictType())



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Exceptions common to OpenStack projects
"""

import logging

from savanna.openstack.common.gettextutils import _

_FATAL_EXCEPTION_FORMAT_ERRORS = False


class Error(Exception):
    def __init__(self, message=None):
        super(Error, self).__init__(message)


class ApiError(Error):
    def __init__(self, message='Unknown', code='Unknown'):
        self.message = message
        self.code = code
        super(ApiError, self).__init__('%s: %s' % (code, message))


class NotFound(Error):
    pass


class UnknownScheme(Error):

    msg = "Unknown scheme '%s' found in URI"

    def __init__(self, scheme):
        msg = self.__class__.msg % scheme
        super(UnknownScheme, self).__init__(msg)


class BadStoreUri(Error):

    msg = "The Store URI %s was malformed. Reason: %s"

    def __init__(self, uri, reason):
        msg = self.__class__.msg % (uri, reason)
        super(BadStoreUri, self).__init__(msg)


class Duplicate(Error):
    pass


class NotAuthorized(Error):
    pass


class NotEmpty(Error):
    pass


class Invalid(Error):
    pass


class BadInputError(Exception):
    """Error resulting from a client sending bad input to a server"""
    pass


class MissingArgumentError(Error):
    pass


class DatabaseMigrationError(Error):
    pass


class ClientConnectionError(Exception):
    """Error resulting from a client connecting to a server"""
    pass


def wrap_exception(f):
    def _wrap(*args, **kw):
        try:
            return f(*args, **kw)
        except Exception as e:
            if not isinstance(e, Error):
                #exc_type, exc_value, exc_traceback = sys.exc_info()
                logging.exception(_('Uncaught exception'))
                #logging.error(traceback.extract_stack(exc_traceback))
                raise Error(str(e))
            raise
    _wrap.func_name = f.func_name
    return _wrap


class OpenstackException(Exception):
    """Base Exception class.

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.
    """
    message = "An unknown exception occurred"

    def __init__(self, **kwargs):
        try:
            self._error_string = self.message % kwargs

        except Exception:
            if _FATAL_EXCEPTION_FORMAT_ERRORS:
                raise
            else:
                # at least get the core message out if something happened
                self._error_string = self.message

    def __str__(self):
        return self._error_string


class MalformedRequestBody(OpenstackException):
    message = "Malformed message body: %(reason)s"


class InvalidContentType(OpenstackException):
    message = "Invalid content type %(content_type)s"



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2013 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Multiple DB API backend support.

Supported configuration options:

The following two parameters are in the 'database' group:
`backend`: DB backend name or full module path to DB backend module.
`use_tpool`: Enable thread pooling of DB API calls.

A DB backend module should implement a method named 'get_backend' which
takes no arguments.  The method can return any object that implements DB
API methods.

*NOTE*: There are bugs in eventlet when using tpool combined with
threading locks. The python logging module happens to use such locks.  To
work around this issue, be sure to specify thread=False with
eventlet.monkey_patch().

A bug for eventlet has been filed here:

https://bitbucket.org/eventlet/eventlet/issue/137/
"""
import functools

from oslo.config import cfg

from savanna.openstack.common import importutils
from savanna.openstack.common import lockutils


db_opts = [
    cfg.StrOpt('backend',
               default='sqlalchemy',
               deprecated_name='db_backend',
               deprecated_group='DEFAULT',
               help='The backend to use for db'),
    cfg.BoolOpt('use_tpool',
                default=False,
                deprecated_name='dbapi_use_tpool',
                deprecated_group='DEFAULT',
                help='Enable the experimental use of thread pooling for '
                     'all DB API calls')
]

CONF = cfg.CONF
CONF.register_opts(db_opts, 'database')


class DBAPI(object):
    def __init__(self, backend_mapping=None):
        if backend_mapping is None:
            backend_mapping = {}
        self.__backend = None
        self.__backend_mapping = backend_mapping

    @lockutils.synchronized('dbapi_backend', 'savanna-')
    def __get_backend(self):
        """Get the actual backend.  May be a module or an instance of
        a class.  Doesn't matter to us.  We do this synchronized as it's
        possible multiple greenthreads started very quickly trying to do
        DB calls and eventlet can switch threads before self.__backend gets
        assigned.
        """
        if self.__backend:
            # Another thread assigned it
            return self.__backend
        backend_name = CONF.database.backend
        self.__use_tpool = CONF.database.use_tpool
        if self.__use_tpool:
            from eventlet import tpool
            self.__tpool = tpool
        # Import the untranslated name if we don't have a
        # mapping.
        backend_path = self.__backend_mapping.get(backend_name,
                                                  backend_name)
        backend_mod = importutils.import_module(backend_path)
        self.__backend = backend_mod.get_backend()
        return self.__backend

    def __getattr__(self, key):
        backend = self.__backend or self.__get_backend()
        attr = getattr(backend, key)
        if not self.__use_tpool or not hasattr(attr, '__call__'):
            return attr

        def tpool_wrapper(*args, **kwargs):
            return self.__tpool.execute(attr, *args, **kwargs)

        functools.update_wrapper(tpool_wrapper, attr)
        return tpool_wrapper



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from novaclient.v1_1 import client as nova_client

from savanna import context
import savanna.utils.openstack.base as base
from savanna.utils.openstack import images
from savanna.utils.openstack import keypairs


def client():
    headers = context.current().headers
    username = headers['X-User-Name']
    token = headers['X-Auth-Token']
    tenant = headers['X-Tenant-Id']
    compute_url = base.url_for(headers, 'compute')

    nova = nova_client.Client(username, token, tenant,
                              auth_url=compute_url)

    nova.client.auth_token = token
    nova.client.management_url = compute_url
    nova.images = images.SavannaImageManager(nova)
    if not hasattr(nova.keypairs, 'get'):
        nova.keypairs = keypairs.SavannaKeypairManager(nova)

    return nova


def get_flavors():
    return [flavor.name for flavor in client().flavors.list()]


def get_flavor(**kwargs):
    return client().flavors.find(**kwargs)


def get_images():
    return [image.id for image in client().images.list()]


def get_limits():
    limits = client().limits.get().absolute
    return dict((l.name, l.value) for l in limits)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from novaclient import exceptions as nova_ex
from oslo.config import cfg
import sqlalchemy as sa
from sqlalchemy.orm import relationship

from savanna.db import model_base as mb
from savanna.utils import configs
from savanna.utils import crypto
from savanna.utils.openstack import nova
from savanna.utils import remote
from savanna.utils import sqlatypes as st


CONF = cfg.CONF
CONF.import_opt('node_domain', 'savanna.service.networks')


## Base mixins for clusters, node groups and their templates

class NodeGroupMixin(mb.IdMixin):
    """Base NodeGroup mixin, add it to subclass is smth like node group."""

    name = sa.Column(sa.String(80), nullable=False)
    flavor_id = sa.Column(sa.String(36), nullable=False)
    image_id = sa.Column(sa.String(36))
    node_processes = sa.Column(st.JsonListType())
    node_configs = sa.Column(st.JsonDictType())
    volumes_per_node = sa.Column(sa.Integer)
    volumes_size = sa.Column(sa.Integer)
    volume_mount_prefix = sa.Column(sa.String(80))


class ClusterMixin(mb.IdMixin, mb.TenantMixin,
                   mb.PluginSpecificMixin, mb.ExtraMixin, mb.DescriptionMixin):
    """Base Cluster mixin, add it to subclass is like cluster object."""

    name = sa.Column(sa.String(80), nullable=False)
    cluster_configs = sa.Column(st.JsonDictType())
    default_image_id = sa.Column(sa.String(36))
    anti_affinity = sa.Column(st.JsonListType())


## Main objects: Cluster, NodeGroup, Instance

class Cluster(mb.SavannaBase, ClusterMixin):
    """Contains all info about cluster."""

    __filter_cols__ = ['private_key']

    __table_args__ = (
        sa.UniqueConstraint('name', 'tenant_id'),
    )

    status = sa.Column(sa.String(80))
    status_description = sa.Column(sa.String(200))

    private_key = sa.Column(sa.Text, default=crypto.generate_private_key())
    user_keypair_id = sa.Column(sa.String(80))

    node_groups = relationship('NodeGroup', cascade="all,delete",
                               backref='cluster')

    info = sa.Column(st.JsonDictType())

    cluster_template_id = sa.Column(sa.String(36),
                                    sa.ForeignKey('ClusterTemplate.id'))
    cluster_template = relationship('ClusterTemplate',
                                    backref="clusters")

    def __init__(self, name, tenant_id, plugin_name, hadoop_version,
                 default_image_id=None, cluster_configs=None,
                 cluster_template_id=None, user_keypair_id=None,
                 anti_affinity=None, description=None):
        self.name = name
        self.tenant_id = tenant_id
        self.plugin_name = plugin_name
        self.hadoop_version = hadoop_version
        self.default_image_id = default_image_id
        self.cluster_configs = cluster_configs or {}
        self.cluster_template_id = cluster_template_id
        self.user_keypair_id = user_keypair_id
        self.anti_affinity = anti_affinity or []
        self.description = description
        self.info = {}

    def to_dict(self):
        d = super(Cluster, self).to_dict()
        d['node_groups'] = [ng.dict for ng in self.node_groups]
        return d

    @property
    def user_keypair(self):
        """Extract user keypair object from nova.

        It contains 'public_key' and 'fingerprint' fields.
        """
        if not hasattr(self, '_user_kp'):
            try:
                self._user_kp = nova.client().keypairs.get(
                    self.user_keypair_id)
            except nova_ex.NotFound:
                self._user_kp = None
        return self._user_kp


class NodeGroup(mb.SavannaBase, NodeGroupMixin):
    """Specifies group of nodes within a cluster."""

    __filter_cols__ = ['id', 'cluster_id']

    __table_args__ = (
        sa.UniqueConstraint('name', 'cluster_id'),
    )

    count = sa.Column(sa.Integer, nullable=False)
    instances = relationship('Instance', cascade="all,delete",
                             backref='node_group',
                             order_by="Instance.instance_name")

    cluster_id = sa.Column(sa.String(36), sa.ForeignKey('Cluster.id'))

    node_group_template_id = sa.Column(sa.String(36),
                                       sa.ForeignKey(
                                           'NodeGroupTemplate.id'))
    node_group_template = relationship('NodeGroupTemplate',
                                       backref="node_groups")

    def __init__(self, name, flavor_id, node_processes, count, image_id=None,
                 node_configs=None, volumes_per_node=0, volumes_size=10,
                 volume_mount_prefix='/volumes/disk',
                 node_group_template_id=None):
        self.name = name
        self.flavor_id = flavor_id
        self.image_id = image_id
        self.node_processes = node_processes
        self.node_configs = node_configs or {}
        self.volumes_per_node = volumes_per_node
        self.volumes_size = volumes_size
        self.volume_mount_prefix = volume_mount_prefix

        self.node_group_template_id = node_group_template_id
        self.count = count

    def get_image_id(self):
        return self.image_id or self.cluster.default_image_id

    @property
    def username(self):
        if not hasattr(self, '_username'):
            image_id = self.get_image_id()
            self._username = nova.client().images.get(image_id).username
        return self._username

    @property
    def configuration(self):
        if hasattr(self, '_all_configs'):
            return self._all_configs

        self._all_configs = configs.merge_configs(
            self.cluster.cluster_configs,
            self.node_configs
        )

        return self._all_configs

    @property
    def storage_paths(self):
        mp = []
        for idx in xrange(1, self.volumes_per_node + 1):
            mp.append(self.volume_mount_prefix + str(idx))

        # Here we assume that NG will use ephemeral drive if no volumes
        if not mp:
            mp = ['/mnt']

        return mp

    def to_dict(self):
        d = super(NodeGroup, self).to_dict()
        d['instances'] = [i.dict for i in self.instances]

        return d


class Instance(mb.SavannaBase, mb.ExtraMixin):
    """An OpenStack instance created for the cluster."""

    __filter_cols__ = ['node_group_id']

    __table_args__ = (
        sa.UniqueConstraint('instance_id', 'node_group_id'),
    )

    node_group_id = sa.Column(sa.String(36), sa.ForeignKey('NodeGroup.id'))
    instance_id = sa.Column(sa.String(36), primary_key=True)
    instance_name = sa.Column(sa.String(80), nullable=False)
    internal_ip = sa.Column(sa.String(15))
    management_ip = sa.Column(sa.String(15))
    volumes = sa.Column(st.JsonListType())

    def __init__(self, node_group_id, instance_id, instance_name,
                 volumes=None):
        self.node_group_id = node_group_id
        self.instance_id = instance_id
        self.instance_name = instance_name
        self.volumes = volumes or []

    @property
    def nova_info(self):
        """Returns info from nova about instance."""
        return nova.client().servers.get(self.instance_id)

    @property
    def username(self):
        return self.node_group.username

    @property
    def hostname(self):
        return self.instance_name

    @property
    def fqdn(self):
        return self.instance_name + '.' + CONF.node_domain

    @property
    def remote(self):
        return remote.InstanceInteropHelper(self)


## Template objects: ClusterTemplate, NodeGroupTemplate, TemplatesRelation

class ClusterTemplate(mb.SavannaBase, ClusterMixin):
    """Template for Cluster."""

    __table_args__ = (
        sa.UniqueConstraint('name', 'tenant_id'),
    )

    node_groups = relationship('TemplatesRelation', cascade="all,delete",
                               backref='cluster_template')

    def __init__(self, name, tenant_id, plugin_name, hadoop_version,
                 default_image_id=None, cluster_configs=None,
                 description=None, anti_affinity=None):
        self.name = name
        self.tenant_id = tenant_id
        self.plugin_name = plugin_name
        self.hadoop_version = hadoop_version
        self.default_image_id = default_image_id
        self.cluster_configs = cluster_configs or {}
        self.anti_affinity = anti_affinity or []
        self.description = description

    def to_dict(self):
        d = super(ClusterTemplate, self).to_dict()
        d['node_groups'] = [tr.dict for tr in
                            self.node_groups]
        return d

    def to_cluster(self, values):
        return Cluster(
            name=values.pop('name', None) or self.name,
            tenant_id=values.pop('tenant_id'),
            plugin_name=values.pop('plugin_name', None) or self.plugin_name,
            hadoop_version=(values.pop('hadoop_version', None)
                            or self.hadoop_version),
            default_image_id=(values.pop('default_image_id')
                              or self.default_image_id),
            cluster_configs=configs.merge_configs(
                self.cluster_configs, values.pop('cluster_configs', None)),
            cluster_template_id=self.id,
            anti_affinity=(values.pop('anti_affinity', None)
                           or self.anti_affinity),
            description=values.pop('description', None),
            **values)


class NodeGroupTemplate(mb.SavannaBase, mb.TenantMixin, mb.PluginSpecificMixin,
                        NodeGroupMixin, mb.DescriptionMixin):
    """Template for NodeGroup."""

    __table_args__ = (
        sa.UniqueConstraint('name', 'tenant_id'),
    )

    def __init__(self, name, tenant_id, flavor_id, plugin_name, hadoop_version,
                 node_processes, image_id=None, node_configs=None,
                 volumes_per_node=0, volumes_size=10,
                 volume_mount_prefix='/volumes/disk', description=None):
        self.name = name
        self.flavor_id = flavor_id
        self.image_id = image_id
        self.node_processes = node_processes
        self.node_configs = node_configs or {}
        self.volumes_per_node = volumes_per_node
        self.volumes_size = volumes_size
        self.volume_mount_prefix = volume_mount_prefix

        self.tenant_id = tenant_id
        self.plugin_name = plugin_name
        self.hadoop_version = hadoop_version
        self.description = description

    def to_object(self, values, cls):
        values.pop('node_group_template_id', None)
        return cls(
            name=values.pop('name', None) or self.name,
            flavor_id=values.pop('flavor_id', None) or self.flavor_id,
            image_id=values.pop('image_id', None) or self.image_id,
            node_processes=(values.pop('node_processes', None)
                            or self.node_processes),
            node_configs=configs.merge_configs(
                self.node_configs, values.pop('node_configs', None)),
            volumes_per_node=(values.pop('volumes_per_node', None)
                              or self.volumes_per_node),
            volumes_size=(values.pop('volumes_size', None)
                          or self.volumes_size),
            volume_mount_prefix=(values.pop('volume_mount_prefix', None)
                                 or self.volume_mount_prefix),
            node_group_template_id=self.id, **values)


class TemplatesRelation(mb.SavannaBase, NodeGroupMixin):
    """NodeGroupTemplate - ClusterTemplate relationship.

    In fact, it's a template of NodeGroup in Cluster.
    """

    __filter_cols__ = ['cluster_template_id', 'created', 'updated', 'id']

    count = sa.Column(sa.Integer, nullable=False)

    cluster_template_id = sa.Column(sa.String(36),
                                    sa.ForeignKey('ClusterTemplate.id'))

    node_group_template_id = sa.Column(sa.String(36),
                                       sa.ForeignKey(
                                           'NodeGroupTemplate.id'))
    node_group_template = relationship('NodeGroupTemplate',
                                       backref="templates_relations")

    def __init__(self, name, flavor_id, node_processes, count, image_id=None,
                 node_configs=None, volumes_per_node=0, volumes_size=10,
                 volume_mount_prefix='/volumes/disk',
                 node_group_template_id=None):
        self.name = name
        self.flavor_id = flavor_id
        self.image_id = image_id
        self.node_processes = node_processes
        self.node_configs = node_configs or {}
        self.volumes_per_node = volumes_per_node
        self.volumes_size = volumes_size
        self.volume_mount_prefix = volume_mount_prefix

        self.node_group_template_id = node_group_template_id
        self.count = count



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import telnetlib

from savanna.tests.integration import base


class ClusterFromClusterTemplateCrudTest(base.ITestCase):

    def setUp(self):
        super(ClusterFromClusterTemplateCrudTest, self).setUp()

        telnetlib.Telnet(self.host, self.port)

        self.create_node_group_templates()

    def crud_clstr_cltr_tmpl(self, node_list):
        cl_tmpl_id = ''
        try:
            cl_tmpl_body = self.make_cluster_template('cl-tmpl', node_list)
            cl_tmpl_id = self.get_object_id(
                'cluster_template', self.post_object(self.url_cl_tmpl,
                                                     cl_tmpl_body, 202))
            clstr_body = self.make_cl_body_cluster_template(cl_tmpl_id)
            self.crud_object(clstr_body, self.url_cluster)
        except Exception as e:
            self.fail('fail: ' + str(e))
        finally:
            self.del_object(self.url_cl_tmpl_with_slash, cl_tmpl_id, 204)

    def test_cluster_nnttdn_jt(self):
        node_list = {self.id_nn_tt_dn: 1, self.id_jt: 1}
        self.crud_clstr_cltr_tmpl(node_list)

    def test_cluster_jtttdn_nn(self):
        node_list = {self.id_jt_tt_dn: 1, self.id_nn: 1}
        self.crud_clstr_cltr_tmpl(node_list)

    def tearDown(self):
        self.delete_node_group_templates()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna import exceptions as ex
from savanna.service import api
import savanna.service.validations.base as b

NODE_GROUP_TEMPLATE_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "minLength": 1,
            "maxLength": 50,
            "format": "valid_name",
        },
        "flavor_id": {
            'type': 'flavor',
        },
        "plugin_name": {
            "type": "string",
        },
        "hadoop_version": {
            "type": "string",
        },
        "node_processes": {
            "type": "array",
            "items": {
                "type": "string",
            },
            "minItems": 1
        },
        "image_id": {
            "type": "string",
            "format": "uuid",
        },
        "node_configs": {
            "type": "configs",
        },
        "volumes_per_node": {
            "type": "integer",
            "minimum": 0,
        },
        "volumes_size": {
            "type": "integer",
            "minimum": 1,
        },
        "volume_mount_prefix": {
            "type": "string",
            "format": "posix_path",
        },
        "description": {
            "type": "string",
        },
    },
    "additionalProperties": False,
    "required": [
        "name",
        "flavor_id",
        "plugin_name",
        "hadoop_version",
        "node_processes",
    ]
}


def check_node_group_template_create(data, **kwargs):
    b.check_node_group_template_unique_name(data['name'])
    b.check_plugin_name_exists(data['plugin_name'])
    b.check_plugin_supports_version(data['plugin_name'],
                                    data['hadoop_version'])
    b.check_node_group_basic_fields(data['plugin_name'],
                                    data['hadoop_version'], data)


def check_node_group_template_usage(node_group_template_id, **kwargs):
    node_groups = []

    for cluster in api.get_clusters():
        node_groups += cluster.node_groups

    for cluster_template in api.get_cluster_templates():
        node_groups += cluster_template.node_groups

    node_group_template_ids = set([node_group.node_group_template_id
                                   for node_group in node_groups])

    if node_group_template_id in node_group_template_ids:
        raise ex.InvalidException(
            "Node group template %s is use" % node_group_template_id)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from alembic import context
from logging import config as logging_config
from savanna.openstack.common import importutils
from sqlalchemy import create_engine, pool

from savanna.db import model_base


importutils.import_module('savanna.db.models')

config = context.config
savanna_config = config.savanna_config

logging_config.fileConfig(config.config_file_name)

# set the target for 'autogenerate' support
target_metadata = model_base.SavannaBase.metadata


def run_migrations_offline():
    context.configure(url=savanna_config.database.connection)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    engine = create_engine(savanna_config.database.connection,
                           poolclass=pool.NullPool)

    connection = engine.connect()
    context.configure(
        connection=connection,
        target_metadata=target_metadata
    )

    try:
        with context.begin_transaction():
            context.run_migrations()
    finally:
        connection.close()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cinderclient.v1 import client as cinder_client

from savanna import context
from savanna.utils.openstack import base


def client():
    headers = context.current().headers
    username = headers['X-User-Name']
    token = headers['X-Auth-Token']
    tenant = headers['X-Tenant-Id']
    volume_url = base.url_for(headers, 'volume')

    cinder = cinder_client.Client(username, token, tenant, volume_url)

    cinder.client.auth_token = token
    cinder.client.management_url = volume_url

    return cinder


def get_volumes():
    return [volume.id for volume in client().volumes.list()]


def get_volume(volume_id):
    return client().volumes.get(volume_id)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest2

from savanna.utils import patches as p
from savanna.utils import xmlutils as x


class XMLUtilsTestCase(unittest2.TestCase):

    def setUp(self):
        p.patch_minidom_writexml()

    def test_load_xml_defaults(self):
        self.assertListEqual(
            [{'name': u'name1', 'value': u'value1', 'description': 'descr1'},
             {'name': u'name2', 'value': u'value2', 'description': 'descr2'},
             {'name': u'name3', 'value': '', 'description': 'descr3'},
             {'name': u'name4', 'value': '', 'description': 'descr4'},
             {'name': u'name5', 'value': u'value5', 'description': ''}],
            x.load_hadoop_xml_defaults(
                'tests/unit/resources/test-default.xml'))

    def test_adjust_description(self):
        self.assertEquals(x._adjust_description("\n"), "")
        self.assertEquals(x._adjust_description("\n  "), "")
        self.assertEquals(x._adjust_description(u"abc\n  def\n  "), "abcdef")
        self.assertEquals(x._adjust_description("abc d\n e f\n"), "abc de f")
        self.assertEquals(x._adjust_description("a\tb\t\nc"), "abc")

    def test_create_hadoop_xml(self):
        conf = x.load_hadoop_xml_defaults(
            'tests/unit/resources/test-default.xml')
        self.assertEquals(x.create_hadoop_xml({'name1': 'some_val1',
                                               'name2': 2}, conf),
                          """<?xml version="1.0" ?>
<?xml-stylesheet type="text/xsl" href="configuration.xsl"?>
<configuration>
  <property>
    <name>name2</name>
    <value>2</value>
  </property>
  <property>
    <name>name1</name>
    <value>some_val1</value>
  </property>
</configuration>
""")



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sqlalchemy as sa

from savanna import context
from savanna.db import model_base as mb
from savanna.tests.unit import base


class TestModel(mb.SavannaBase, mb.IdMixin):
    test_field = sa.Column(sa.String(80))


def _insert_test_object():
    t = TestModel()
    t.test_field = 123
    context.model_save(t)

    return t


class ModelHelpersTest(base.DbTestCase):
    def test_model_save(self):
        self.assertEqual(0, len(context.model_query(TestModel).all()))

        t = _insert_test_object()

        self.assertEqual(1, len(context.model_query(TestModel).all()))

        db_t = context.model_query(TestModel).first()

        self.assertEqual(t.id, db_t.id)
        self.assertEqual(t.test_field, db_t.test_field)

    def test_model_update(self):
        _insert_test_object()

        t = context.current().session.query(TestModel).first()

        context.model_update(t, test_field=42)

        db_t = context.model_query(TestModel).first()

        self.assertEqual(t.id, db_t.id)
        self.assertEqual(42, db_t.test_field)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import telnetlib

from savanna.tests.integration import base


class ClusterNodeProcessesCrudTest(base.ITestCase):

    def setUp(self):
        super(ClusterNodeProcessesCrudTest, self).setUp()

        telnetlib.Telnet(self.host, self.port)

    def test_crud_cluster_nn(self):
        """This test checks cluster creation with topology | NN | via
        node process.
        """
        node_processes = {'NN': 1}
        body = self.make_cl_body_node_processes(node_processes)
        self.crud_object(body, self.url_cluster)

    def test_crud_cluster_nn_dn(self):
        """This test checks cluster creation with topology | NN | DN | via
        node processes.
        """
        node_processes = {'NN': 1, 'DN': 2}
        body = self.make_cl_body_node_processes(node_processes)
        self.crud_object(body, self.url_cluster)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna import context as ctx
import savanna.db.models as m


def _model_filter_by(model, **kwargs):
    return ctx.model_query(model, project_only=True).filter_by(**kwargs)


## Cluster ops

def get_clusters(**args):
    return _model_filter_by(m.Cluster, **args).all()


def get_cluster(**args):
    return _model_filter_by(m.Cluster, **args).first()


def create_cluster(values):
    values = values.copy()
    session = ctx.current().session
    with session.begin():
        values['tenant_id'] = ctx.current().tenant_id
        ngs_vals = values.pop('node_groups', [])
        cluster_tmpl_id = values.pop('cluster_template_id', None)
        if cluster_tmpl_id:
            cluster_tmpl = get_cluster_template(id=cluster_tmpl_id)
            cluster = cluster_tmpl.to_cluster(values)
        else:
            cluster = m.Cluster(**values)

        if not ngs_vals and cluster_tmpl_id:
            # copy node groups from cluster template
            ngs_vals = cluster_tmpl.dict['node_groups']

        for ng in ngs_vals:
            tmpl_id = ng.get('node_group_template_id')
            if tmpl_id:
                tmpl = get_node_group_template(id=tmpl_id)
                node_group = tmpl.to_object(ng, m.NodeGroup)
            else:
                node_group = m.NodeGroup(**ng)
            cluster.node_groups.append(node_group)
            session.add(node_group)
        session.add(cluster)

        return cluster


def terminate_cluster(cluster):
    with ctx.current().session.begin():
        ctx.current().session.delete(cluster)


## ClusterTemplate ops

def get_cluster_templates(**args):
    return _model_filter_by(m.ClusterTemplate, **args).all()


def get_cluster_template(**args):
    return _model_filter_by(m.ClusterTemplate, **args).first()


def create_cluster_template(values):
    values = values.copy()
    session = ctx.current().session
    with session.begin():
        values['tenant_id'] = ctx.current().tenant_id
        ngts_vals = values.pop('node_groups', [])
        cluster_template = m.ClusterTemplate(**values)
        for ngt in ngts_vals:
            tmpl_id = ngt.get('node_group_template_id')
            if tmpl_id:
                tmpl = get_node_group_template(id=tmpl_id)
                node_group = tmpl.to_object(ngt, m.TemplatesRelation)
                node_group.cluster_template_id = cluster_template.id
            else:
                node_group = m.TemplatesRelation(**ngt)
            cluster_template.node_groups.append(node_group)
            session.add(node_group)
        session.add(cluster_template)

        return cluster_template


def persist_cluster_template(cluster_template):
    session = ctx.current().session
    with session.begin():
        session.add(cluster_template)

    return cluster_template


def terminate_cluster_template(**args):
    with ctx.current().session.begin():
        ctx.current().session.delete(get_cluster_template(**args))


## NodeGroupTemplate ops

def get_node_group_templates(**args):
    return _model_filter_by(m.NodeGroupTemplate, **args).all()


def get_node_group_template(**args):
    return _model_filter_by(m.NodeGroupTemplate, **args).first()


def create_node_group_template(values):
    values = values.copy()
    session = ctx.current().session
    with session.begin():
        values['tenant_id'] = ctx.current().tenant_id
        node_group_template = m.NodeGroupTemplate(**values)
        session.add(node_group_template)
        return node_group_template


def terminate_node_group_template(**args):
    with ctx.current().session.begin():
        ctx.current().session.delete(get_node_group_template(**args))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import unittest2

from savanna import context
import savanna.swift.swift_helper as h

GENERAL_PREFIX = "fs.swift."
SERVICE_PREFIX = "service.savanna."

GENERAL = ["impl", "connect.timeout", "socket.timeout",
           "connect.retry.count", "connect.throttle.delay",
           "blocksize", "partsize", "requestsize"]

SERVICE_SPECIFIC = ["auth.url", "tenant",
                    "username", "password", "http.port",
                    "https.port", "public", "location-aware",
                    "region", "apikey"]


class SwiftIntegrationTestCase(unittest2.TestCase):
    def setUp(self):
        context.set_ctx(
            context.Context('test_user', 'test_tenant', 'test_auth_token',
                            {'X-Tenant-Name': "test_tenant"}))

    @mock.patch('savanna.swift.swift_helper._retrieve_auth_url')
    def test_get_swift_configs(self, authUrlConfig):
        authUrlConfig.return_value = "http://localhost:8080/v2.0/tokens"

        result = h.get_swift_configs()
        self.assertEqual(7, len(result))
        self.assertIn({'name': "fs.swift.service.savanna.location-aware",
                       'value': 'true', 'description': ''}, result)
        self.assertIn({'name': "fs.swift.service.savanna.tenant",
                       'value': 'test_tenant', 'description': ''}, result)
        self.assertIn({'name': "fs.swift.service.savanna.http.port",
                       'value': '8080', 'description': ''}, result)



# Copyright (c) 2013 Hortonworks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import savanna.db.models as m
from savanna.plugins import provisioning as p


def convert(cluster_template, normalized_config):
    cluster_template.hadoop_version = normalized_config.hadoop_version
    for ng in normalized_config.node_groups:
        ct_node_group = m.NodeGroup(ng.name, ng.flavor,
                                    ng.node_processes,
                                    ng.count)
        cluster_template.node_groups.append(ct_node_group)
    for entry in normalized_config.cluster_configs:
        ci = entry.config
        ct_config = {"config": p.Config(ci.name, ci.applicable_target,
                     ci.scope, config_type=ci.type,
                     default_value=ci.default_value,
                     is_optional=ci.is_optional,
                     description=ci.description),
                     "value": entry.value}
        cluster_template.cluster_configs.append(ct_config)

    return cluster_template



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import json
import os
import telnetlib
import time

import keystoneclient.v2_0
import requests
import unittest2

import savanna.tests.integration.configs.parameters as param
from savanna.utils import remote


def swift_test(fun):
    if param.ENABLE_SWIFT_TESTS:
        return fun

    return unittest2.skip("Swift tests has been disabled")(fun)


class ITestCase(unittest2.TestCase):

    def setUp(self):
        self.port = param.SAVANNA_PORT
        self.host = param.SAVANNA_HOST

        self.baseurl = 'http://' + self.host + ':' + self.port

        self.keystone = keystoneclient.v2_0.Client(
            username=param.OS_USERNAME,
            password=param.OS_PASSWORD,
            tenant_name=param.OS_TENANT_NAME,
            auth_url=param.OS_AUTH_URL
        )

        self.tenant = self.keystone.tenant_id
        self.token = self.keystone.auth_token

        self.url_version = '/'

        self.url_ngt = '/v1.0/%s/node-group-templates' % self.tenant
        self.url_ngt_with_slash = '/v1.0/%s/node-group-templates/'\
                                  % self.tenant

        self.url_cluster = '/v1.0/%s/clusters' % self.tenant
        self.url_cluster_with_slash = '/v1.0/%s/clusters/' % self.tenant

        self.url_cl_tmpl = '/v1.0/%s/cluster-templates' % self.tenant
        self.url_cl_tmpl_with_slash = '/v1.0/%s/cluster-templates/'\
                                      % self.tenant

        self.url_plugins = '/v1.0/%s/plugins' % self.tenant
        self.url_plugins_with_slash = '/v1.0/%s/plugins/' % self.tenant

        self.url_images = '/v1.0/%s/images' % self.tenant
        self.url_images_with_slash = '/v1.0/%s/images/' % self.tenant

#----------------------methods_for_CRUD_operations-----------------------------

    def _post(self, url, body):
        URL = self.baseurl + url
        resp = requests.post(URL, data=body, headers={
            'x-auth-token': self.token, 'Content-Type': 'application/json'})
        data = json.loads(resp.content) if resp.status_code == 202 \
            else resp.content
        print('URL = %s \n data = %s \n response = %s \n data = %s \n'
              % (URL, body, resp.status_code, data))
        return resp

    def _put(self, url, body):
        URL = self.baseurl + url
        resp = requests.put(URL, data=body, headers={
            'x-auth-token': self.token, 'Content-Type': 'application/json'})
        data = json.loads(resp.content)
        print('URL = %s \n data = %s \n response = %s \n data = %s \n'
              % (URL, body, resp.status_code, data))
        return resp

    def _get(self, url, printing):
        URL = self.baseurl + url
        resp = requests.get(URL, headers={'x-auth-token': self.token})
        if printing:
            print('URL = %s \n response = %s \n' % (URL, resp.status_code))
        if resp.status_code != 200:
            data = json.loads(resp.content)
            print('data= %s \n' % data)
        return resp

    def _delete(self, url):
        URL = self.baseurl + url
        resp = requests.delete(URL, headers={'x-auth-token': self.token})
        print('URL = %s \n response = %s \n' % (URL, resp.status_code))
        if resp.status_code != 204:
            data = json.loads(resp.content)
            print('data= %s \n' % data)
        return resp

    def post_object(self, url, body, code):
        post = self._post(url, json.dumps(body))
        self.assertEquals(post.status_code, code)
        data = json.loads(post.content)
        return data

    def put_object(self, url, object_id, body, code):
        data = self._put(url + object_id, json.dumps(body))
        self.assertEquals(data.status_code, code)
        data = json.loads(data.content)
        return data

    def get_object(self, url, obj_id, code, printing=False):
        rv = self._get(url + obj_id, printing)
        self.assertEquals(rv.status_code, code)
        data = json.loads(rv.content)
        return data

    def del_object(self, url, obj_id, code):
        rv = self._delete(url + obj_id)
        self.assertEquals(rv.status_code, code)
        if rv.status_code != 204:
            data = json.loads(rv.content)
            return data
        else:
            return rv.status_code

    def crud_object(self, body, url):
        data = self.post_object(url, body, 202)
        get_url = None
        object_id = None
        crud_object = None
        try:
            if url == self.url_cluster:
                crud_object = 'cluster'
                get_url = self.url_cluster_with_slash
            elif url == self.url_ngt:
                crud_object = 'node_group_template'
                get_url = self.url_ngt_with_slash
            else:
                crud_object = 'cluster_template'
                get_url = self.url_cl_tmpl_with_slash
            data = data[crud_object]
            object_id = data.get('id')
            if crud_object == 'cluster':
                self.await_cluster_active(object_id)
        except Exception as e:
            self.fail('failure: ' + str(e))
        finally:
            self.del_object(get_url, object_id, 204)
            if crud_object == 'cluster':
                time.sleep(5)
        return object_id

#----------------------make_different_template_body----------------------------

    def make_node_group_template(self, gr_name, desc, n_proc):
        processes = ['tasktracker', 'datanode']
        if n_proc == 'JT+NN':
            processes = ['jobtracker', 'namenode']
        elif n_proc == 'JT':
            processes = ['jobtracker']
        elif n_proc == 'NN':
            processes = ['namenode']
        elif n_proc == 'TT':
            processes = ['tasktracker']
        elif n_proc == 'DN':
            processes = ['datanode']
        elif n_proc == 'NN+TT+DN':
            processes = ['namenode', 'tasktracker', 'datanode']
        elif n_proc == 'JT+TT+DN':
            processes = ['jobtracker', 'tasktracker', 'datanode']
        group_template = dict(
            name=gr_name,
            description=desc,
            flavor_id=param.FLAVOR_ID,
            plugin_name=param.PLUGIN_NAME,
            hadoop_version=param.HADOOP_VERSION,
            node_processes=processes,
            node_configs={
                'HDFS': {},
                'MapReduce': {}
            }
        )
        return group_template

    def make_cluster_template(self, name, ngt_list):
        cluster_template = dict(
            name=name,
            plugin_name=param.PLUGIN_NAME,
            hadoop_version=param.HADOOP_VERSION,
            cluster_configs={},
            node_groups=[]
        )
        for key, value in ngt_list.items():
            ngt = dict(
                name='',
                node_group_template_id='',
                count=1
            )
            ngt['node_group_template_id'] = key
            ngt['count'] = value
            data = self.get_object(self.url_ngt_with_slash, key, 200)
            name = data['node_group_template']['name']
            ngt['name'] = name
            cluster_template['node_groups'].append(ngt)
        return cluster_template

    def make_cl_body_cluster_template(self, cl_tmpl_id):
        cluster_body = dict(
            name=param.CLUSTER_NAME_CRUD,
            plugin_name=param.PLUGIN_NAME,
            hadoop_version=param.HADOOP_VERSION,
            cluster_template_id=cl_tmpl_id,
            default_image_id=param.IMAGE_ID,
            user_keypair_id=param.SSH_KEY
        )
        return cluster_body

    def make_cl_body_node_processes(self, node_processes):
        cluster_body = dict(
            name=param.CLUSTER_NAME_CRUD,
            plugin_name=param.PLUGIN_NAME,
            hadoop_version=param.HADOOP_VERSION,
            user_keypair_id=param.SSH_KEY,
            default_image_id=param.IMAGE_ID,
            cluster_configs={},
            node_groups=[]
        )
        for key, value in node_processes.items():
            processes = ['jobtracker', 'namenode']
            ng_name = 'jt-nn'
            if key == 'TT+DN':
                processes = ['tasktracker', 'datanode']
                ng_name = 'tt-dn'
            elif key == 'JT':
                processes = ['jobtracker']
                ng_name = 'jt'
            elif key == 'NN':
                processes = ['namenode']
                ng_name = 'nn'
            elif key == 'TT':
                processes = ['tasktracker']
                ng_name = 'tt'
            elif key == 'DN':
                processes = ['datanode']
                ng_name = 'dn'
            elif key == 'JT+TT+DN':
                processes = ['jobtracker', 'tasktracker', 'datanode']
                ng_name = 'jt-tt-dn'
            elif key == 'NN+TT+DN':
                processes = ['namenode', 'tasktracker', 'datanode']
                ng_name = 'nn-tt-dn'
            cluster_body['node_groups'].append(dict(
                name=ng_name,
                flavor_id=param.FLAVOR_ID,
                node_processes=processes,
                count=value
            ))
        return cluster_body

    def make_cl_body_node_group_templates(self, ngt_id_list):
        cluster_body = dict(
            name=param.CLUSTER_NAME_CRUD,
            plugin_name=param.PLUGIN_NAME,
            hadoop_version=param.HADOOP_VERSION,
            user_keypair_id=param.SSH_KEY,
            default_image_id=param.IMAGE_ID,
            cluster_configs={},
            node_groups=[]
        )
        for key, value in ngt_id_list.items():
            data = self.get_object(self.url_ngt_with_slash, key, 200)
            name = data['node_group_template']['name']
            cluster_body['node_groups'].append(dict(
                name=name,
                node_group_template_id=key,
                count=value
            ))
        return cluster_body

    def make_cl_node_processes_ngt(self, node_processes, ngt_id_list):
        cluster_body = self.make_cl_body_node_processes(node_processes)
        for key, value in ngt_id_list.items():
            data = self.get_object(self.url_ngt_with_slash, key, 200)
            name = data['node_group_template']['name']
            cluster_body['node_groups'].append(dict(
                name=name,
                node_group_template_id=key,
                count=value
            ))
        return cluster_body

#-------------------make_node_group_template_and_get_id------------------------

    def create_node_group_templates(self):
        self.id_tt = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'worker-tt', 'qa probe', 'TT'), 202))
        self.id_jt = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'master-jt', 'qa probe', 'JT'), 202))

        self.id_nn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'master-nn', 'qa probe', 'NN'), 202))

        self.id_dn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'worker-dn', 'qa probe', 'DN'), 202))

        self.id_tt_dn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'worker-tt-dn', 'qa probe', 'TT+DN'), 202))

        self.id_jt_nn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'master-jt-nn', 'qa probe', 'JT+NN'), 202))

        self.id_nn_tt_dn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'nn-tt-dn', 'qa probe', 'NN+TT+DN'), 202))

        self.id_jt_tt_dn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'jt-tt-dn', 'qa probe', 'JT+TT+DN'), 202))

#---------------------delete_node_group_template-------------------------------

    def delete_node_group_templates(self):
        self.del_object(self.url_ngt_with_slash, self.id_jt_nn, 204)
        self.del_object(self.url_ngt_with_slash, self.id_jt, 204)
        self.del_object(self.url_ngt_with_slash, self.id_nn, 204)
        self.del_object(self.url_ngt_with_slash, self.id_tt, 204)
        self.del_object(self.url_ngt_with_slash, self.id_dn, 204)
        self.del_object(self.url_ngt_with_slash, self.id_tt_dn, 204)
        self.del_object(self.url_ngt_with_slash, self.id_nn_tt_dn, 204)
        self.del_object(self.url_ngt_with_slash, self.id_jt_tt_dn, 204)

#------------------------------helper_methods----------------------------------

    def await_cluster_active(self, object_id):
        get_data = self.get_object(self.url_cluster_with_slash, object_id, 200)
        get_data = get_data['cluster']
        i = 1
        while get_data['status'] != 'Active':
            print('GET_STATUS: ', get_data['status'])
            if i > param.TIMEOUT * 6:
                print('\n Data for cluster: ' + str(get_data) + '\n')
                self.fail(
                    'cluster is not getting status \'Active\', '
                    'passed %d minutes' % param.TIMEOUT)
            get_data = self.get_object(
                self.url_cluster_with_slash, object_id, 200)
            get_data = get_data['cluster']
            time.sleep(10)
            i += 1

    def get_object_id(self, obj, body):
        data = body[obj]
        return data['id']

    def ssh_connection(self, host):
        return remote.setup_ssh_connection(host, param.NODE_USERNAME,
                                           open(param.PATH_TO_SSH).read())

    def execute_command(self, host, cmd):
        with contextlib.closing(self.ssh_connection(host)) as ssh:
            return remote.execute_command(ssh, cmd)

    def write_file_to(self, host, remote_file, data):
        with contextlib.closing(self.ssh_connection(host)) as ssh:
            return remote.write_file_to(ssh.open_sftp(), remote_file, data)

    def read_file_from(self, host, remote_file):
        with contextlib.closing(self.ssh_connection(host)) as ssh:
            return remote.read_file_from(ssh.open_sftp(), remote_file)

    def transfer_script_to_node(self, host,
                                script='hadoop_test/hadoop_test_script.sh'):
        self.write_file_to(str(host),
                           'script.sh',
                           open('%s/integration/%s' % (os.getcwd(),
                                                       script)).read())
        self.execute_command(str(host), 'chmod 777 script.sh')

    def try_telnet(self, host, port):
        try:
            telnetlib.Telnet(host, port)

        except Exception as e:
            self.fail('Telnet has failed: ' + str(e) +
                      '\n    NODE_IP: %s, PORT: %s' % (host, port))

    def create_cluster_and_get_id(self, cluster_body):
        data = self.post_object(self.url_cluster, cluster_body, 202)
        cluster_id = data['cluster']['id']

        self.await_cluster_active(cluster_id)

        return cluster_id

    def create_cluster_using_ngt_and_get_id(self, cl_tmpl_id, name):
        try:
            clstr_body = self.make_cl_body_cluster_template(cl_tmpl_id)
            clstr_body['name'] = name
            return self.create_cluster_and_get_id(clstr_body)

        except Exception as e:
            self.del_object(self.url_cl_tmpl_with_slash, cl_tmpl_id, 204)
            self.fail('failure: ' + str(e))

    def get_instances_ip_and_node_processes_list(self, cluster_id):
        get_data = self.get_object(
            self.url_cluster_with_slash, cluster_id, 200, True)
        node_groups = get_data['cluster']['node_groups']

        instances_ip = {}

        for node_group in node_groups:
            instances = node_group['instances']
            for instance in instances:
                management_ip = instance['management_ip']
                instances_ip[management_ip] = \
                    node_group['node_processes']

        return instances_ip

    def get_namenode_ip_and_tt_dn_count(self, instances_ip):
        tasktracker_count = 0
        datanode_count = 0
        node_count = 0

        namenode_ip = ''

        portmap = {
            'jobtracker': param.JT_PORT,
            'namenode': param.NN_PORT,
            'tasktracker': param.TT_PORT,
            'datanode': param.DN_PORT
        }

        for host, processes in instances_ip.items():
            self.try_telnet(host, '22')
            node_count += 1

            for process in processes:
                self.try_telnet(host, portmap[process])

            if 'tasktracker' in processes:
                tasktracker_count += 1

            if 'datanode' in processes:
                datanode_count += 1

            if 'namenode' in processes:
                namenode_ip = host

        return {
            'namenode_ip': namenode_ip,
            'tasktracker_count': tasktracker_count,
            'datanode_count': datanode_count,
            'node_count': node_count
        }

    def await_active_workers_for_namenode(self, data):
        attempts_count = 20

        while True:
            active_tasktrackers_count = int(
                self.execute_command(data['namenode_ip'],
                                     'sudo su -c "cd %s && hadoop job \
                                     -list-active-trackers" hadoop | wc -l'
                                     % param.HADOOP_DIRECTORY)[1])

            active_datanodes_count = int(
                self.execute_command(data['namenode_ip'],
                                     'sudo su -c "hadoop dfsadmin -report" \
                                     hadoop | grep "Datanodes available:.*" | \
                                     awk \'{print $3}\'')[1])

            if active_tasktrackers_count == \
                    data['tasktracker_count'] and active_datanodes_count == \
                    data['datanode_count']:
                break

            if attempts_count == 0:
                self.fail('tasktracker or datanode cannot be started '
                          'within 1 minute.')

            time.sleep(3)

            attempts_count -= 1



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import telnetlib

from savanna.tests.integration import base


class ClusterTemplatesCrudTest(base.ITestCase):

    def setUp(self):
        super(ClusterTemplatesCrudTest, self).setUp()

        telnetlib.Telnet(self.host, self.port)

        self.create_node_group_templates()

    def test_crud_cl_tmpl_nnjt_dntt(self):
        """This test checks cluster template creation with
        configuration | JT + NN | TT + DN |.
        """
        node_list = {self.id_jt_nn: 1, self.id_tt_dn: 2}
        cl_tmpl_body = self.make_cluster_template('cl-tmpl-1', node_list)
        self.crud_object(cl_tmpl_body, self.url_cl_tmpl)

    def test_crud_cl_tmpl_nn_jt_dn_tt(self):
        """This test checks cluster template creation with
        configuration | JT | NN | TT | DN |.
        """
        node_list = {self.id_nn: 1, self.id_jt: 1, self.id_tt: 5,
                     self.id_dn: 5}
        cl_tmpl_body = self.make_cluster_template('cl-tmpl-2', node_list)
        self.crud_object(cl_tmpl_body, self.url_cl_tmpl)

    def test_crud_cl_tmpl_nn_jt_dntt(self):
        """This test checks cluster template creation with
        configuration | JT | NN | TT + DN |.
        """
        node_list = {self.id_jt: 1, self.id_nn: 1, self.id_tt_dn: 7}
        cl_tmpl_body = self.make_cluster_template('cl-tmpl-3', node_list)
        self.crud_object(cl_tmpl_body, self.url_cl_tmpl)

    def test_crud_cl_tmpl_nn(self):
        """This test checks cluster template creation with
        configuration | NN |.
        """
        node_list = {self.id_nn: 1}
        cl_tmpl_body = self.make_cluster_template('cl-tmpl-4', node_list)
        self.crud_object(cl_tmpl_body, self.url_cl_tmpl)

    def test_crud_cl_tmpl_nn_dn(self):
        """This test checks cluster template creation with
        configuration | NN | DN |.
        """
        node_list = {self.id_nn: 1, self.id_dn: 7}
        cl_tmpl_body = self.make_cluster_template('cl-tmpl-5', node_list)
        self.crud_object(cl_tmpl_body, self.url_cl_tmpl)

    def test_crud_cl_tmpl_nn_jt_dn(self):
        """This test checks cluster template creation with
        configuration | JT | NN | DN |.
        """
        node_list = {self.id_nn: 1, self.id_jt: 1, self.id_dn: 7}
        cl_tmpl_body = self.make_cluster_template('cl-tmpl-6', node_list)
        self.crud_object(cl_tmpl_body, self.url_cl_tmpl)

    def test_crud_cl_tmpl_jt_nnttdn(self):
        """This test checks cluster template creation with
        configuration | JT | NN + TT + DN |.
        """
        node_list = {self.id_jt: 1, self.id_nn_tt_dn: 1}
        cl_tmpl_body = self.make_cluster_template('cl-tmpl-7', node_list)
        self.crud_object(cl_tmpl_body, self.url_cl_tmpl)

    def test_crud_cl_tmpl_nn_jtttdn(self):
        """This test checks cluster template creation with
        configuration | NN | JT + TT + DN |.
        """
        node_list = {self.id_nn: 1, self.id_jt_tt_dn: 1}
        cl_tmpl_body = self.make_cluster_template('cl-tmpl-8', node_list)
        self.crud_object(cl_tmpl_body, self.url_cl_tmpl)

    def test_crud_cl_tmpl_nnttdn_jt_ttdn(self):
        """This test checks cluster template creation with
        configuration | JT | NN + TT + DN | TT + DN |.
        """
        node_list = {self.id_nn_tt_dn: 1, self.id_jt: 1, self.id_tt_dn: 7}
        cl_tmpl_body = self.make_cluster_template('cl-tmpl-9', node_list)
        self.crud_object(cl_tmpl_body, self.url_cl_tmpl)

    def test_crud_cl_tmpl_jtttdn_nn_ttdn(self):
        """This test checks cluster template creation with
        configuration | JT + TT + DN | NN | TT + DN |.
        """
        node_list = {self.id_jt_tt_dn: 1, self.id_nn: 1, self.id_tt_dn: 7}
        cl_tmpl_body = self.make_cluster_template('cl-tmpl-10', node_list)
        self.crud_object(cl_tmpl_body, self.url_cl_tmpl)

    def tearDown(self):
        self.delete_node_group_templates()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import telnetlib

from savanna.tests.integration import base


class ClusterNodeProcessesNodeGroupTemplatesCrudTest(base.ITestCase):

    def setUp(self):
        super(ClusterNodeProcessesNodeGroupTemplatesCrudTest, self).setUp()

        telnetlib.Telnet(self.host, self.port)

        self.create_node_group_templates()

    def test_crud_cluster_nn_jt_dn(self):
        """This test checks cluster creation with topology | JT | NN | DN | via
        node processes and node group template. Node processes are JT
        and NN.
        """
        node_processes = {'JT': 1, 'NN': 1}
        ngt_id_list = {self.id_dn: 2}
        body = self.make_cl_node_processes_ngt(node_processes, ngt_id_list)
        self.crud_object(body, self.url_cluster)

    def test_crud_cluster_nn_jt_ttdn(self):
        """This test checks cluster creation with topology | JT | NN | TT+DN |
        via node processes and node group template. Node processes are
        JT and NN.
        """
        node_processes = {'JT': 1, 'NN': 1}
        ngt_id_list = {self.id_tt_dn: 2}
        body = self.make_cl_node_processes_ngt(node_processes, ngt_id_list)
        self.crud_object(body, self.url_cluster)

    def tearDown(self):
        self.delete_node_group_templates()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest2

from savanna import context as ctx
import savanna.db.models as m
from savanna.tests.unit import base as models_test_base


SAMPLE_CONFIGS = {
    'a': 'av',
    'b': 123,
    'c': [1, '2', u"3"]
}


class TemplatesModelTest(models_test_base.DbTestCase):
    def testCreateNodeGroupTemplate(self):
        session = ctx.current().session
        with session.begin():
            ngt = m.NodeGroupTemplate('ngt-1', 't-1', 'f-1', 'p-1', 'hv-1',
                                      ['np-1', 'np-2'],
                                      node_configs=SAMPLE_CONFIGS)
            session.add(ngt)

        res = session.query(m.NodeGroupTemplate).filter_by().first()

        self.assertIsValidModelObject(res)
        self.assertEquals(['np-1', 'np-2'], res.node_processes)
        self.assertEquals(SAMPLE_CONFIGS, res.node_configs)

        res_dict = self.get_clean_dict(res)

        self.assertEqual(res_dict, {
            'flavor_id': 'f-1',
            'hadoop_version': 'hv-1',
            'name': 'ngt-1',
            'node_configs': SAMPLE_CONFIGS,
            'node_processes': ['np-1', 'np-2'],
            'plugin_name': 'p-1',
            'image_id': None,
            'volume_mount_prefix': '/volumes/disk',
            'volumes_per_node': 0,
            'volumes_size': 10,
            'description': None
        })

    def testCreateClusterTemplate(self):
        session = ctx.current().session
        with session.begin():
            c = m.ClusterTemplate('c-1', 't-1', 'p-1', 'hv-1',
                                  cluster_configs=SAMPLE_CONFIGS)
            session.add(c)

        res = session.query(m.ClusterTemplate).filter_by().first()
        self.assertIsValidModelObject(res)
        self.assertEqual(SAMPLE_CONFIGS, res.cluster_configs)

        res_dict = self.get_clean_dict(res)

        self.assertEqual(res_dict, {
            'cluster_configs': SAMPLE_CONFIGS,
            'hadoop_version': 'hv-1',
            'name': 'c-1',
            'plugin_name': 'p-1',
            'node_groups': [],
            'default_image_id': None,
            'description': None,
            'anti_affinity': []
        })

    @unittest2.skip('add_node_group_template has been removed')
    def testCreateClusterTemplateWithNodeGroupTemplates(self):
        session = ctx.current().session
        with session.begin():
            ct = m.ClusterTemplate('ct', 't-1', 'p-1', 'hv-1')
            session.add(ct)

            ngts = []
            for i in xrange(0, 3):
                ngt = m.NodeGroupTemplate('ngt-%s' % i, 't-1', 'f-1', 'p-1',
                                          'hv-1', ['np-1', 'np-2'])
                session.add(ngt)
                session.flush()
                rel = ct.add_node_group_template({
                    'node_group_template_id': ngt.id,
                    'name': 'group-%s' % i,
                    'count': 5 + i
                })
                session.add(rel)
                ngts.append(ngt)

        with session.begin():
            res = session.query(m.ClusterTemplate).filter_by().first()
            self.assertIsValidModelObject(res)

            self.assertEqual(len(res.node_group_templates), 3)
            self.assertEqual(set(t.name for t in res.node_group_templates),
                             set('ngt-%s' % i for i in xrange(0, 3)))


class NestedTemplateTest(unittest2.TestCase):
    @unittest2.skip('template to obj conversions are not fully implemented')
    def test_nested_templates(self):
        ct = m.ClusterTemplate('ct', 't-1', 'p-1', 'hv-1')
        ct.cluster_configs = {
            "service:map-reduce": {
                "ct": 0,
                "nt": 0,
                "c": 0,
                "n": 0
            },
            "service:hdfs": {
                "ct": 0,
                "nt": 0,
                "c": 0,
                "n": 0
            }
        }

        c = m.Cluster('c-1', 't-1', 'p-1', 'hv-1')
        c.cluster_configs = {
            "service:map-reduce": {
                "c": 1,
                "nt": 1,
                "n": 1
            },
            "service:hdfs": {
                "c": 1,
                "nt": 1,
                "n": 1
            }
        }
        c.base_cluster_template = ct

        ngt = m.NodeGroupTemplate("ngt", "t-1", "f-1", "p-1",
                                  "h-1", ["tt", "dn"])
        ngt.node_configs = {
            "service:map-reduce": {
                "nt": 2,
                "n": 2
            },
            "service:hdfs": {
                "nt": 2,
                "n": 2
            }
        }

        ng = m.NodeGroup("ng", "f-i", ["tt", "dn"], 1)
        ng.cluster = c
        ng.base_node_group_template = ngt
        ng.node_configs = {
            "service:map-reduce": {
                "n": 3
            },
            "service:hdfs": {
                "n": 3
            }
        }

        self.assertEqual(ng.configuration["service:map-reduce"]["ct"], 0)
        self.assertEqual(ng.configuration["service:map-reduce"]["c"], 1)
        self.assertEqual(ng.configuration["service:map-reduce"]["nt"], 2)
        self.assertEqual(ng.configuration["service:map-reduce"]["n"], 3)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2011 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Only (de)serialization utils hasn't been removed to decrease requirements
# number.

"""Utility methods for working with WSGI servers."""

import datetime
from xml.dom import minidom
from xml.parsers import expat

from savanna.openstack.common import exception
from savanna.openstack.common.gettextutils import _
from savanna.openstack.common import jsonutils
from savanna.openstack.common import log as logging
from savanna.openstack.common import xmlutils

LOG = logging.getLogger(__name__)


class ActionDispatcher(object):
    """Maps method name to local methods through action name."""

    def dispatch(self, *args, **kwargs):
        """Find and call local method."""
        action = kwargs.pop('action', 'default')
        action_method = getattr(self, str(action), self.default)
        return action_method(*args, **kwargs)

    def default(self, data):
        raise NotImplementedError()


class DictSerializer(ActionDispatcher):
    """Default request body serialization"""

    def serialize(self, data, action='default'):
        return self.dispatch(data, action=action)

    def default(self, data):
        return ""


class JSONDictSerializer(DictSerializer):
    """Default JSON request body serialization"""

    def default(self, data):
        def sanitizer(obj):
            if isinstance(obj, datetime.datetime):
                _dtime = obj - datetime.timedelta(microseconds=obj.microsecond)
                return _dtime.isoformat()
            return unicode(obj)
        return jsonutils.dumps(data, default=sanitizer)


class XMLDictSerializer(DictSerializer):

    def __init__(self, metadata=None, xmlns=None):
        """
        :param metadata: information needed to deserialize xml into
                         a dictionary.
        :param xmlns: XML namespace to include with serialized xml
        """
        super(XMLDictSerializer, self).__init__()
        self.metadata = metadata or {}
        self.xmlns = xmlns

    def default(self, data):
        # We expect data to contain a single key which is the XML root.
        root_key = data.keys()[0]
        doc = minidom.Document()
        node = self._to_xml_node(doc, self.metadata, root_key, data[root_key])

        return self.to_xml_string(node)

    def to_xml_string(self, node, has_atom=False):
        self._add_xmlns(node, has_atom)
        return node.toprettyxml(indent='    ', encoding='UTF-8')

    #NOTE (ameade): the has_atom should be removed after all of the
    # xml serializers and view builders have been updated to the current
    # spec that required all responses include the xmlns:atom, the has_atom
    # flag is to prevent current tests from breaking
    def _add_xmlns(self, node, has_atom=False):
        if self.xmlns is not None:
            node.setAttribute('xmlns', self.xmlns)
        if has_atom:
            node.setAttribute('xmlns:atom', "http://www.w3.org/2005/Atom")

    def _to_xml_node(self, doc, metadata, nodename, data):
        """Recursive method to convert data members to XML nodes."""
        result = doc.createElement(nodename)

        # Set the xml namespace if one is specified
        # TODO(justinsb): We could also use prefixes on the keys
        xmlns = metadata.get('xmlns', None)
        if xmlns:
            result.setAttribute('xmlns', xmlns)

        #TODO(bcwaldon): accomplish this without a type-check
        if type(data) is list:
            collections = metadata.get('list_collections', {})
            if nodename in collections:
                metadata = collections[nodename]
                for item in data:
                    node = doc.createElement(metadata['item_name'])
                    node.setAttribute(metadata['item_key'], str(item))
                    result.appendChild(node)
                return result
            singular = metadata.get('plurals', {}).get(nodename, None)
            if singular is None:
                if nodename.endswith('s'):
                    singular = nodename[:-1]
                else:
                    singular = 'item'
            for item in data:
                node = self._to_xml_node(doc, metadata, singular, item)
                result.appendChild(node)
        #TODO(bcwaldon): accomplish this without a type-check
        elif type(data) is dict:
            collections = metadata.get('dict_collections', {})
            if nodename in collections:
                metadata = collections[nodename]
                for k, v in data.items():
                    node = doc.createElement(metadata['item_name'])
                    node.setAttribute(metadata['item_key'], str(k))
                    text = doc.createTextNode(str(v))
                    node.appendChild(text)
                    result.appendChild(node)
                return result
            attrs = metadata.get('attributes', {}).get(nodename, {})
            for k, v in data.items():
                if k in attrs:
                    result.setAttribute(k, str(v))
                else:
                    node = self._to_xml_node(doc, metadata, k, v)
                    result.appendChild(node)
        else:
            # Type is atom
            node = doc.createTextNode(str(data))
            result.appendChild(node)
        return result

    def _create_link_nodes(self, xml_doc, links):
        link_nodes = []
        for link in links:
            link_node = xml_doc.createElement('atom:link')
            link_node.setAttribute('rel', link['rel'])
            link_node.setAttribute('href', link['href'])
            if 'type' in link:
                link_node.setAttribute('type', link['type'])
            link_nodes.append(link_node)
        return link_nodes


class TextDeserializer(ActionDispatcher):
    """Default request body deserialization"""

    def deserialize(self, datastring, action='default'):
        return self.dispatch(datastring, action=action)

    def default(self, datastring):
        return {}


class JSONDeserializer(TextDeserializer):

    def _from_json(self, datastring):
        try:
            return jsonutils.loads(datastring)
        except ValueError:
            msg = _("cannot understand JSON")
            raise exception.MalformedRequestBody(reason=msg)

    def default(self, datastring):
        return {'body': self._from_json(datastring)}


class XMLDeserializer(TextDeserializer):

    def __init__(self, metadata=None):
        """
        :param metadata: information needed to deserialize xml into
                         a dictionary.
        """
        super(XMLDeserializer, self).__init__()
        self.metadata = metadata or {}

    def _from_xml(self, datastring):
        plurals = set(self.metadata.get('plurals', {}))

        try:
            node = xmlutils.safe_minidom_parse_string(datastring).childNodes[0]
            return {node.nodeName: self._from_xml_node(node, plurals)}
        except expat.ExpatError:
            msg = _("cannot understand XML")
            raise exception.MalformedRequestBody(reason=msg)

    def _from_xml_node(self, node, listnames):
        """Convert a minidom node to a simple Python type.

        :param listnames: list of XML node names whose subnodes should
                          be considered list items.

        """

        if len(node.childNodes) == 1 and node.childNodes[0].nodeType == 3:
            return node.childNodes[0].nodeValue
        elif node.nodeName in listnames:
            return [self._from_xml_node(n, listnames) for n in node.childNodes]
        else:
            result = dict()
            for attr in node.attributes.keys():
                result[attr] = node.attributes[attr].nodeValue
            for child in node.childNodes:
                if child.nodeType != node.TEXT_NODE:
                    result[child.nodeName] = self._from_xml_node(child,
                                                                 listnames)
            return result

    def find_first_child_named(self, parent, name):
        """Search a nodes children for the first child with a given name"""
        for node in parent.childNodes:
            if node.nodeName == name:
                return node
        return None

    def find_children_named(self, parent, name):
        """Return all of a nodes children who have the given name"""
        for node in parent.childNodes:
            if node.nodeName == name:
                yield node

    def extract_text(self, node):
        """Get the text field contained by the given node"""
        if len(node.childNodes) == 1:
            child = node.childNodes[0]
            if child.nodeType == child.TEXT_NODE:
                return child.nodeValue
        return ""

    def default(self, datastring):
        return {'body': self._from_xml(datastring)}



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

import paramiko

from savanna.utils import crypto


def setup_ssh_connection(host, username, private_key):
    """Setup SSH connection to the host using username and private key."""
    if type(private_key) in [str, unicode]:
        private_key = crypto.to_paramiko_private_key(private_key)
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(host, username=username, pkey=private_key)

    return ssh


def execute_command(ssh_connection, cmd):
    """Execute specified command remotely using existing ssh connection.

    Return exit code and stdout data of the executed command.
    """
    chan = ssh_connection.get_transport().open_session()
    chan.exec_command(cmd)
    retcode = chan.recv_exit_status()
    buf = ''
    while chan.recv_ready():
        buf += chan.recv(1024)
    return retcode, buf


def write_file_to(sftp, remote_file, data):
    """Create remote file using existing ssh connection and write the given
    data to it.
    """
    fl = sftp.file(remote_file, 'w')
    fl.write(data)
    fl.close()


def write_files_to(sftp, files):
    """Copy file->data dictionary in a single ssh connection.
    """
    for fl, data in files.iteritems():
        write_file_to(sftp, fl, data)


def read_file_from(sftp, remote_file):
    """Read remote file from the specified host and return given data."""
    fl = sftp.file(remote_file, 'r')
    data = fl.read()
    fl.close()
    return data


def replace_remote_string(ssh_connection, remote_file, old_str, new_str):
    """Replaces strings in remote file using sed command."""
    cmd = "sudo sed -i 's,%s,%s,g' %s" % (old_str, new_str, remote_file)
    return execute_command(ssh_connection, cmd)


class InstanceInteropHelper(object):
    def __init__(self, instance):
        self.instance = instance

    def __enter__(self):
        self.bulk = BulkInstanceInteropHelper(self)
        return self.bulk

    def __exit__(self, *exc_info):
        self.bulk.close()

    def ssh_connection(self):
        return setup_ssh_connection(
            self.instance.management_ip, self.instance.username,
            self.instance.node_group.cluster.private_key)

    def execute_command(self, cmd):
        with contextlib.closing(self.ssh_connection()) as ssh:
            return execute_command(ssh, cmd)

    def write_file_to(self, remote_file, data):
        with contextlib.closing(self.ssh_connection()) as ssh:
            return write_file_to(ssh.open_sftp(), remote_file, data)

    def write_files_to(self, files):
        with contextlib.closing(self.ssh_connection()) as ssh:
            return write_files_to(ssh.open_sftp(), files)

    def read_file_from(self, remote_file):
        with contextlib.closing(self.ssh_connection()) as ssh:
            return read_file_from(ssh.open_sftp(), remote_file)

    def replace_remote_string(self, remote_file, old_str, new_str):
        with contextlib.closing(self.ssh_connection()) as ssh:
            return replace_remote_string(ssh, remote_file, old_str, new_str)


class BulkInstanceInteropHelper(object):
    def __init__(self, helper):
        self.helper = helper
        self._ssh = None
        self._sftp = None

    def close(self):
        if self._sftp:
            self._sftp.close()
        if self._ssh:
            self._ssh.close()

    def ssh_connection(self):
        if not self._ssh:
            self._ssh = self.helper.ssh_connection()
        return self._ssh

    def sftp_connection(self):
        if not self._sftp:
            self._sftp = self.ssh_connection().open_sftp()
        return self._sftp

    def execute_command(self, cmd):
        return execute_command(self.ssh_connection(), cmd)

    def write_file_to(self, remote_file, data):
        return write_file_to(self.sftp_connection(), remote_file, data)

    def write_files_to(self, files):
        return write_files_to(self.sftp_connection(), files)

    def read_file_from(self, remote_file):
        return read_file_from(self.sftp_connection(), remote_file)

    def replace_remote_string(self, remote_file, old_str, new_str):
        return replace_remote_string(self.ssh_connection(), remote_file,
                                     old_str, new_str)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

'''
JSON related utilities.

This module provides a few things:

    1) A handy function for getting an object down to something that can be
    JSON serialized.  See to_primitive().

    2) Wrappers around loads() and dumps().  The dumps() wrapper will
    automatically use to_primitive() for you if needed.

    3) This sets up anyjson to use the loads() and dumps() wrappers if anyjson
    is available.
'''


import datetime
import functools
import inspect
import itertools
import json
import types
import xmlrpclib

import netaddr
import six

from savanna.openstack.common import timeutils


_nasty_type_tests = [inspect.ismodule, inspect.isclass, inspect.ismethod,
                     inspect.isfunction, inspect.isgeneratorfunction,
                     inspect.isgenerator, inspect.istraceback, inspect.isframe,
                     inspect.iscode, inspect.isbuiltin, inspect.isroutine,
                     inspect.isabstract]

_simple_types = (types.NoneType, int, basestring, bool, float, long)


def to_primitive(value, convert_instances=False, convert_datetime=True,
                 level=0, max_depth=3):
    """Convert a complex object into primitives.

    Handy for JSON serialization. We can optionally handle instances,
    but since this is a recursive function, we could have cyclical
    data structures.

    To handle cyclical data structures we could track the actual objects
    visited in a set, but not all objects are hashable. Instead we just
    track the depth of the object inspections and don't go too deep.

    Therefore, convert_instances=True is lossy ... be aware.

    """
    # handle obvious types first - order of basic types determined by running
    # full tests on nova project, resulting in the following counts:
    # 572754 <type 'NoneType'>
    # 460353 <type 'int'>
    # 379632 <type 'unicode'>
    # 274610 <type 'str'>
    # 199918 <type 'dict'>
    # 114200 <type 'datetime.datetime'>
    #  51817 <type 'bool'>
    #  26164 <type 'list'>
    #   6491 <type 'float'>
    #    283 <type 'tuple'>
    #     19 <type 'long'>
    if isinstance(value, _simple_types):
        return value

    if isinstance(value, datetime.datetime):
        if convert_datetime:
            return timeutils.strtime(value)
        else:
            return value

    # value of itertools.count doesn't get caught by nasty_type_tests
    # and results in infinite loop when list(value) is called.
    if type(value) == itertools.count:
        return six.text_type(value)

    # FIXME(vish): Workaround for LP bug 852095. Without this workaround,
    #              tests that raise an exception in a mocked method that
    #              has a @wrap_exception with a notifier will fail. If
    #              we up the dependency to 0.5.4 (when it is released) we
    #              can remove this workaround.
    if getattr(value, '__module__', None) == 'mox':
        return 'mock'

    if level > max_depth:
        return '?'

    # The try block may not be necessary after the class check above,
    # but just in case ...
    try:
        recursive = functools.partial(to_primitive,
                                      convert_instances=convert_instances,
                                      convert_datetime=convert_datetime,
                                      level=level,
                                      max_depth=max_depth)
        if isinstance(value, dict):
            return dict((k, recursive(v)) for k, v in value.iteritems())
        elif isinstance(value, (list, tuple)):
            return [recursive(lv) for lv in value]

        # It's not clear why xmlrpclib created their own DateTime type, but
        # for our purposes, make it a datetime type which is explicitly
        # handled
        if isinstance(value, xmlrpclib.DateTime):
            value = datetime.datetime(*tuple(value.timetuple())[:6])

        if convert_datetime and isinstance(value, datetime.datetime):
            return timeutils.strtime(value)
        elif hasattr(value, 'iteritems'):
            return recursive(dict(value.iteritems()), level=level + 1)
        elif hasattr(value, '__iter__'):
            return recursive(list(value))
        elif convert_instances and hasattr(value, '__dict__'):
            # Likely an instance of something. Watch for cycles.
            # Ignore class member vars.
            return recursive(value.__dict__, level=level + 1)
        elif isinstance(value, netaddr.IPAddress):
            return six.text_type(value)
        else:
            if any(test(value) for test in _nasty_type_tests):
                return six.text_type(value)
            return value
    except TypeError:
        # Class objects are tricky since they may define something like
        # __iter__ defined but it isn't callable as list().
        return six.text_type(value)


def dumps(value, default=to_primitive, **kwargs):
    return json.dumps(value, default=default, **kwargs)


def loads(s):
    return json.loads(s)


def load(s):
    return json.load(s)


try:
    import anyjson
except ImportError:
    pass
else:
    anyjson._modules.append((__name__, 'dumps', TypeError,
                                       'loads', ValueError, 'load'))
    anyjson.force_implementation(__name__)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import savanna.exceptions as ex
import savanna.plugins.base as plugin_base
import savanna.service.api as api
import savanna.service.validations.base as b
import savanna.service.validations.cluster_templates as cl_t


def _build_node_groups_schema():
    schema = copy.deepcopy(cl_t.CLUSTER_TEMPLATE_SCHEMA)
    return schema['properties']['node_groups']


CLUSTER_SCALING_SCHEMA = {
    "type": "object",
    "properties": {
        "resize_node_groups": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string",
                    },
                    "count": {
                        "type": "integer",
                        "minimum": 0,
                    },
                },
                "additionalProperties": False,
                "required": [
                    "name",
                    "count",
                ]
            },
            "minItems": 1
        },
        "add_node_groups": _build_node_groups_schema(),
    },
    "additionalProperties": False,
    "anyOf": [
        {
            "required": ["resize_node_groups"]
        },
        {
            "required": ["add_node_groups"]
        }
    ]

}


def check_cluster_scaling(data, cluster_id, **kwargs):
    cluster = api.get_cluster(id=cluster_id)
    if not (plugin_base.PLUGINS.is_plugin_implements(cluster.plugin_name,
                                                     'scale_cluster') and (
            plugin_base.PLUGINS.is_plugin_implements(cluster.plugin_name,
                                                     'decommission_nodes'))):
        raise ex.InvalidException(
            "Requested plugin '%s' doesn't support cluster scaling feature"
            % cluster.plugin_name)

    if cluster.status != 'Active':
        raise ex.InvalidException("Cluster cannot be scaled not in 'Active' "
                                  "status. Cluster status: " + cluster.status)

    if data.get("resize_node_groups"):
        b.check_resize(cluster, data['resize_node_groups'])

    if data.get("add_node_groups"):
        b.check_add_node_groups(cluster, data['add_node_groups'])



# Copyright (c) 2013 Hortonworks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from savanna.plugins.hdp import ambariplugin as ap
from savanna.plugins.hdp import clusterspec as cs
import unittest2


class AmbariPluginTest(unittest2.TestCase):
    def test_get_node_processes(self):
        plugin = ap.AmbariPlugin()
        #TODO(jspeidel): provide meaningful input
        service_components = plugin.get_node_processes(1)

        self.assertEqual(5, len(service_components))
        components = service_components['HDFS']
        self.assertIn('NAMENODE', components)
        self.assertIn('DATANODE', components)
        self.assertIn('SECONDARY_NAMENODE', components)
        self.assertIn('HDFS_CLIENT', components)

        components = service_components['MAPREDUCE']
        self.assertIn('JOBTRACKER', components)
        self.assertIn('TASKTRACKER', components)
        self.assertIn('MAPREDUCE_CLIENT', components)

        components = service_components['GANGLIA']
        self.assertIn('GANGLIA_SERVER', components)
        self.assertIn('GANGLIA_MONITOR', components)

        components = service_components['NAGIOS']
        self.assertIn('NAGIOS_SERVER', components)

        components = service_components['AMBARI']
        self.assertIn('AMBARI_SERVER', components)
        self.assertIn('AMBARI_AGENT', components)

    def test_convert(self):
        plugin = ap.AmbariPlugin()
        cluster = TestCluster()
        plugin.convert(cluster,
                       os.path.join(os.path.realpath('../plugins'), 'hdp',
                                    'resources',
                                    'default-cluster.template'))

        with open(os.path.join(os.path.realpath('../plugins'), 'hdp',
                               'resources',
                               'default-cluster.template'), 'r') as f:
            normalized_config = cs.ClusterSpec(f.read()).normalize()

        self.assertEquals(normalized_config.hadoop_version,
                          cluster.hadoop_version)
        self.assertEquals(len(normalized_config.node_groups),
                          len(cluster.node_groups))
        self.assertEquals(len(normalized_config.cluster_configs),
                          len(cluster.cluster_configs))
        #TODO(jspeidel): drill down into node_groups and cluster_configs

    def test_update_infra(self):
        plugin = ap.AmbariPlugin()
        cluster = TestCluster()
        plugin.update_infra(cluster)

        for node_group in cluster.node_groups:
            self.assertEquals(cluster.default_image_id, node_group.image)

    def test_get_configs(self):
        plugin = ap.AmbariPlugin()
        configs = plugin.get_configs("1.1.2")
        self.assertEqual(612, len(configs),
                         "wrong number of configuration properties")


class TestCluster:
    def __init__(self):
        self.hadoop_version = None
        self.cluster_configs = []
        self.node_groups = []
        self.default_image_id = '11111'



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

import eventlet
from eventlet import corolocal

from savanna.db import api as db_api
from savanna.openstack.common import log as logging


LOG = logging.getLogger(__name__)


# TODO(slukjanov): it'll be better to use common_context.RequestContext as base
class Context(object):
    def __init__(self, user_id, tenant_id, auth_token, headers, **kwargs):
        if kwargs:
            LOG.warn('Arguments dropped when creating context: %s', kwargs)

        self.user_id = user_id
        self.tenant_id = tenant_id
        self.auth_token = auth_token
        self.headers = headers
        self._db_session = None

    def clone(self):
        return Context(self.user_id,
                       self.tenant_id,
                       self.auth_token,
                       self.headers)

    @property
    def session(self):
        if self._db_session is None:
            self._db_session = db_api.get_session()
        return self._db_session


_CTXS = threading.local()
_CTXS._curr_ctxs = {}


def has_ctx():
    ident = corolocal.get_ident()
    return ident in _CTXS._curr_ctxs and _CTXS._curr_ctxs[ident]


def ctx():
    if not has_ctx():
        # TODO(slukjanov): replace with specific error
        raise RuntimeError("Context isn't available here")
    return _CTXS._curr_ctxs[corolocal.get_ident()]


def current():
    return ctx()


def session(context=None):
    context = context or ctx()
    return context.session


def set_ctx(new_ctx):
    ident = corolocal.get_ident()

    if not new_ctx and ident in _CTXS._curr_ctxs:
        del _CTXS._curr_ctxs[ident]

    if new_ctx:
        _CTXS._curr_ctxs[ident] = new_ctx


def model_query(model, context=None, project_only=None):
    context = context or ctx()
    query = context.session.query(model)

    if project_only:
        query = query.filter_by(tenant_id=context.tenant_id)

    return query


def model_save(model, context=None):
    context = context or ctx()
    with context.session.begin():
        context.session.add(model)
    return model


def model_update(model, context=None, **kwargs):
    if not hasattr(model, '__table__'):
        # TODO(slikjanov): replace with specific exception
        raise RuntimeError("Specified object isn't model, class: %s"
                           % model.__class__.__name__)
    columns = model.__table__.columns
    for prop in kwargs:
        if prop not in columns:
            # TODO(slukjanov): replace with specific exception
            raise RuntimeError(
                "Model class '%s' doesn't contains specified property '%s'"
                % (model.__class__.__name__, prop))
        setattr(model, prop, kwargs[prop])

    return model_save(model, context)


def spawn(func, *args, **kwargs):
    ctx = current().clone()

    def wrapper(ctx, func, *args, **kwargs):
        set_ctx(ctx)
        func(*args, **kwargs)

    eventlet.spawn(wrapper, ctx, func, *args, **kwargs)


def sleep(seconds=0):
    eventlet.sleep(seconds)
